#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Information about units of measure and mapping between Pyomo function names
and external function names.
"""

__author__ = "John Eslick"

from pyomo.environ import units as pu

external_function_map = {
    # Thermo properties as a function of delta and tau
    "p_func": {  # pressure
        "fname": "p",
        "units": pu.kPa,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
        "doc": "p(comp, delta, tau)",
    },
    "u_func": {  # internal energy
        "fname": "u",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
        "doc": "u(comp, delta, tau)",
    },
    "s_func": {  # entropy
        "fname": "s",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
        "doc": "s(comp, delta, tau)",
    },
    "h_func": {  # enthalpy
        "fname": "h",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
        "doc": "h(comp, delta, tau)",
    },
    "g_func": {  # Gibbs free energy
        "fname": "g",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
        "doc": "g(comp, delta, tau)",
    },
    "f_func": {  # Helmholtz free energy
        "fname": "f",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
        "doc": "f(comp, delta, tau)",
    },
    "cv_func": {  # constant volume heat capacity
        "fname": "cv",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
        "doc": "cv(comp, delta, tau)",
    },
    "cp_func": {  # constant pressure heat capacity
        "fname": "cp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
        "doc": "cp(comp, delta, tau)",
    },
    "w_func": {  # speed of sound
        "fname": "w",
        "units": pu.m / pu.s,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
        "doc": "w(comp, delta, tau)",
    },
    "v_func": {  # specific volume
        "fname": "v",
        "units": pu.m**3 / pu.kg,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
        "doc": "v(comp, delta, tau)",
    },
    "mu_func": {  # viscosity
        "fname": "mu",
        "units": pu.uPa * pu.s,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
        "doc": "mu(comp, delta, tau)",
    },
    "lambda_func": {  # thermal conductivity
        "fname": "lambda",
        "units": pu.mW / pu.m / pu.K,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
        "doc": "lambda(comp, delta, tau)",
    },
    "sigma_func": {  # surface tension
        "fname": "sigma",
        "units": pu.mN / pu.m,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
        "doc": "sigma(comp, delta, tau)",
    },
    # Functions of (h, p)
    "t_hp_func": {  # temperature
        "fname": "T_hp",
        "units": pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "T(comp, enthalpy, pressure)",
    },
    "tau_hp_func": {  # Tc/T
        "fname": "tau_hp",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "tau(comp, enthalpy, pressure)",
    },
    "vf_hp_func": {  # vapor fraction
        "fname": "vf_hp",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "vf(comp, enthalpy, pressure)",
    },
    "u_hp_func": {  # internal energy
        "fname": "u_hp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "u(comp, enthalpy, pressure)",
    },
    "s_hp_func": {  # entropy
        "fname": "s_hp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "s(comp, enthalpy, pressure)",
    },
    "g_hp_func": {  # Gibbs free energy
        "fname": "g_hp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "g(comp, enthalpy, pressure)",
    },
    "f_hp_func": {  # Helmholtz free energy
        "fname": "f_hp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "f(comp, enthalpy, pressure)",
    },
    "cv_hp_func": {  # constant volume heat capacity
        "fname": "cv_hp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "cv(comp, enthalpy, pressure)",
    },
    "cp_hp_func": {  # constant pressure heat capacity
        "fname": "cp_hp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "cp(comp, enthalpy, pressure)",
    },
    "w_hp_func": {  # speed of sound
        "fname": "w_hp",
        "units": pu.m / pu.s,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "w(comp, enthalpy, pressure)",
    },
    "v_hp_func": {  # specific volume
        "fname": "v_hp",
        "units": pu.m**3 / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "v(comp, enthalpy, pressure)",
    },
    "mu_hp_func": {  # viscosity
        "fname": "mu_hp",
        "units": pu.uPa * pu.s,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "mu(comp, enthalpy, pressure)",
    },
    "lambda_hp_func": {  # thermal conductivity
        "fname": "lambda_hp",
        "units": pu.mW / pu.m / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "lambda(comp, enthalpy, pressure)",
    },
    "sigma_hp_func": {  # surface tension
        "fname": "sigma_hp",
        "units": pu.mN / pu.m,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "sigma(comp, enthalpy, pressure)",
    },
    "h_vap_hp_func": {  # vapor enthalpy
        "fname": "h_vap_hp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "h_vap(comp, enthalpy, pressure)",
    },
    "u_vap_hp_func": {  # internal energy
        "fname": "u_vap_hp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "u_vap(comp, enthalpy, pressure)",
    },
    "s_vap_hp_func": {  # entropy
        "fname": "s_vap_hp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "s_vap(comp, enthalpy, pressure)",
    },
    "g_vap_hp_func": {  # Gibbs free energy
        "fname": "g_vap_hp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "g_vap(comp, enthalpy, pressure)",
    },
    "f_vap_hp_func": {  # Helmholtz free energy
        "fname": "f_vap_hp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "f_vap(comp, enthalpy, pressure)",
    },
    "cv_vap_hp_func": {  # constant volume heat capacity
        "fname": "cv_vap_hp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "cv_vap(comp, enthalpy, pressure)",
    },
    "cp_vap_hp_func": {  # constant pressure heat capacity
        "fname": "cp_vap_hp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "cp_vap(comp, enthalpy, pressure)",
    },
    "w_vap_hp_func": {  # speed of sound
        "fname": "w_vap_hp",
        "units": pu.m / pu.s,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "w_vap(comp, enthalpy, pressure)",
    },
    "v_vap_hp_func": {  # specific volume
        "fname": "v_vap_hp",
        "units": pu.m**3 / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "v_vap(comp, enthalpy, pressure)",
    },
    "mu_vap_hp_func": {  # viscosity
        "fname": "mu_vap_hp",
        "units": pu.uPa * pu.s,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "mu_vap(comp, enthalpy, pressure)",
    },
    "lambda_vap_hp_func": {  # thermal conductivity
        "fname": "lambda_vap_hp",
        "units": pu.mW / pu.m / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "lambda_vap(comp, enthalpy, pressure)",
    },
    "sigma_vap_hp_func": {  # surface tension
        "fname": "sigma_vap_hp",
        "units": pu.mN / pu.m,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "sigma_vap(comp, enthalpy, pressure)",
    },
    "h_liq_hp_func": {  # liquid enthalpy
        "fname": "h_liq_hp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "h_liq(comp, enthalpy, pressure)",
    },
    "u_liq_hp_func": {  # internal energy
        "fname": "u_liq_hp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "u_liq(comp, enthalpy, pressure)",
    },
    "s_liq_hp_func": {  # entropy
        "fname": "s_liq_hp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "s_liq(comp, enthalpy, pressure)",
    },
    "g_liq_hp_func": {  # Gibbs free energy
        "fname": "g_liq_hp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "g_liq(comp, enthalpy, pressure)",
    },
    "f_liq_hp_func": {  # Helmholtz free energy
        "fname": "f_liq_hp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "f_liq(comp, enthalpy, pressure)",
    },
    "cv_liq_hp_func": {  # constant volume heat capacity
        "fname": "cv_liq_hp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "cv_liq(comp, enthalpy, pressure)",
    },
    "cp_liq_hp_func": {  # constant pressure heat capacity
        "fname": "cp_liq_hp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "cp_liq(comp, enthalpy, pressure)",
    },
    "w_liq_hp_func": {  # speed of sound
        "fname": "w_liq_hp",
        "units": pu.m / pu.s,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "w_liq(comp, enthalpy, pressure)",
    },
    "v_liq_hp_func": {  # specific volume
        "fname": "v_liq_hp",
        "units": pu.m**3 / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "v_liq(comp, enthalpy, pressure)",
    },
    "mu_liq_hp_func": {  # viscosity
        "fname": "mu_liq_hp",
        "units": pu.uPa * pu.s,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "mu_liq(comp, enthalpy, pressure)",
    },
    "lambda_liq_hp_func": {  # thermal conductivity
        "fname": "lambda_liq_hp",
        "units": pu.mW / pu.m / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "lambda_liq(comp, enthalpy, pressure)",
    },
    "sigma_liq_hp_func": {  # surface tension
        "fname": "sigma_liq_hp",
        "units": pu.mN / pu.m,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "sigma_liq(comp, enthalpy, pressure)",
    },
    # Functions of (s, p)
    "t_sp_func": {  # temperature
        "fname": "T_sp",
        "units": pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "T(comp, entropy, pressure)",
    },
    "tau_sp_func": {  # Tc/T
        "fname": "tau_sp",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "tau(comp, entropy, pressure)",
    },
    "vf_sp_func": {  # vapor fraction
        "fname": "vf_sp",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "vf(comp, entropy, pressure)",
    },
    "u_sp_func": {  # internal energy
        "fname": "u_sp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "u(comp, entropy, pressure)",
    },
    "h_sp_func": {  # enthalpy
        "fname": "h_sp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "h(comp, entropy, pressure)",
    },
    "g_sp_func": {  # Gibbs free energy
        "fname": "g_sp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "g(comp, entropy, pressure)",
    },
    "f_sp_func": {  # Helmholtz free energy
        "fname": "f_sp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "f(comp, entropy, pressure)",
    },
    "cv_sp_func": {  # constant volume heat capacity
        "fname": "cv_sp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "cv(comp, entropy, pressure)",
    },
    "cp_sp_func": {  # constant pressure heat capacity
        "fname": "cp_sp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "cp(comp, entropy, pressure)",
    },
    "w_sp_func": {  # speed of sound
        "fname": "w_sp",
        "units": pu.m / pu.s,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "w(comp, entropy, pressure)",
    },
    "v_sp_func": {  # specific volume
        "fname": "v_sp",
        "units": pu.m**3 / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "v(comp, entropy, pressure)",
    },
    "mu_sp_func": {  # viscosity
        "fname": "mu_sp",
        "units": pu.uPa * pu.s,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "mu(comp, entropy, pressure)",
    },
    "lambda_sp_func": {  # thermal conductivity
        "fname": "lambda_sp",
        "units": pu.mW / pu.m / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "lambda(comp, entropy, pressure)",
    },
    "sigma_sp_func": {  # surface tension
        "fname": "sigma_sp",
        "units": pu.mN / pu.m,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "sigma(comp, entropy, pressure)",
    },
    "s_vap_sp_func": {  # vapor entropy
        "fname": "s_vap_sp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "s_vap(comp, entropy, pressure)",
    },
    "u_vap_sp_func": {  # internal energy
        "fname": "u_vap_sp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "u_vap(comp, entropy, pressure)",
    },
    "h_vap_sp_func": {  # enthalpy
        "fname": "h_vap_sp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "h_vap(comp, entropy, pressure)",
    },
    "g_vap_sp_func": {  # Gibbs free energy
        "fname": "g_vap_sp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "g_vap(comp, entropy, pressure)",
    },
    "f_vap_sp_func": {  # Helmholtz free energy
        "fname": "f_vap_sp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "f_vap(comp, entropy, pressure)",
    },
    "cv_vap_sp_func": {  # constant volume heat capacity
        "fname": "cv_vap_sp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "cv_vap(comp, entropy, pressure)",
    },
    "cp_vap_sp_func": {  # constant pressure heat capacity
        "fname": "cp_vap_sp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "cp_vap(comp, entropy, pressure)",
    },
    "w_vap_sp_func": {  # speed of sound
        "fname": "w_vap_sp",
        "units": pu.m / pu.s,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "w_vap(comp, entropy, pressure)",
    },
    "v_vap_sp_func": {  # specific volume
        "fname": "v_vap_sp",
        "units": pu.m**3 / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "v_vap(comp, entropy, pressure)",
    },
    "mu_vap_sp_func": {  # viscosity
        "fname": "mu_vap_sp",
        "units": pu.uPa * pu.s,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "mu_vap(comp, entropy, pressure)",
    },
    "lambda_vap_sp_func": {  # thermal conductivity
        "fname": "lambda_vap_sp",
        "units": pu.mW / pu.m / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "lambda_vap(comp, entropy, pressure)",
    },
    "sigma_vap_sp_func": {  # surface tension
        "fname": "sigma_vap_sp",
        "units": pu.mN / pu.m,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "sigma_vap(comp, entropy, pressure)",
    },
    "s_liq_sp_func": {  # liquid entropy
        "fname": "s_liq_sp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "s_liq(comp, entropy, pressure)",
    },
    "u_liq_sp_func": {  # internal energy
        "fname": "u_liq_sp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "u_liq(comp, entropy, pressure)",
    },
    "h_liq_sp_func": {  # enthalpy
        "fname": "h_liq_sp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "h_liq(comp, entropy, pressure)",
    },
    "g_liq_sp_func": {  # Gibbs free energy
        "fname": "g_liq_sp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "g_liq(comp, entropy, pressure)",
    },
    "f_liq_sp_func": {  # Helmholtz free energy
        "fname": "f_liq_sp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "f_liq(comp, entropy, pressure)",
    },
    "cv_liq_sp_func": {  # constant volume heat capacity
        "fname": "cv_liq_sp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "cv_liq(comp, entropy, pressure)",
    },
    "cp_liq_sp_func": {  # constant pressure heat capacity
        "fname": "cp_liq_sp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "cp_liq(comp, entropy, pressure)",
    },
    "w_liq_sp_func": {  # speed of sound
        "fname": "w_liq_sp",
        "units": pu.m / pu.s,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "w_liq(comp, entropy, pressure)",
    },
    "v_liq_sp_func": {  # specific volume
        "fname": "v_liq_sp",
        "units": pu.m**3 / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "v_liq(comp, entropy, pressure)",
    },
    "mu_liq_sp_func": {  # viscosity
        "fname": "mu_liq_sp",
        "units": pu.uPa * pu.s,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "mu_liq(comp, entropy, pressure)",
    },
    "lambda_liq_sp_func": {  # thermal conductivity
        "fname": "lambda_liq_sp",
        "units": pu.mW / pu.m / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "lambda_liq(comp, entropy, pressure)",
    },
    "sigma_liq_sp_func": {  # surface tension
        "fname": "sigma_liq_sp",
        "units": pu.mN / pu.m,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg / pu.K, pu.kPa],
        "doc": "sigma_liq(comp, entropy, pressure)",
    },
    # Functions of (u, p)
    "t_up_func": {  # temperature
        "fname": "T_up",
        "units": pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "T(comp, internal energy, pressure)",
    },
    "tau_up_func": {  # Tc/T
        "fname": "tau_up",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "tau(comp, internal energy, pressure)",
    },
    "vf_up_func": {  # vapor fraction
        "fname": "vf_up",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "vf(comp, internal energy, pressure)",
    },
    "h_up_func": {  # enthalpy
        "fname": "h_up",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "h(comp, internal energy, pressure)",
    },
    "s_up_func": {  # entropy
        "fname": "s_up",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "s(comp, internal energy, pressure)",
    },
    "g_up_func": {  # Gibbs free energy
        "fname": "g_up",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "g(comp, internal energy, pressure)",
    },
    "f_up_func": {  # Helmholtz free energy
        "fname": "f_up",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "f(comp, internal energy, pressure)",
    },
    "cv_up_func": {  # constant volume heat capacity
        "fname": "cv_up",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "cv(comp, internal energy, pressure)",
    },
    "cp_up_func": {  # constant pressure heat capacity
        "fname": "cp_up",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "cp(comp, internal energy, pressure)",
    },
    "w_up_func": {  # speed of sound
        "fname": "w_up",
        "units": pu.m / pu.s,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "w(comp, internal energy, pressure)",
    },
    "v_up_func": {  # specific volume
        "fname": "v_up",
        "units": pu.m**3 / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "v(comp, internal energy, pressure)",
    },
    "mu_up_func": {  # viscosity
        "fname": "mu_up",
        "units": pu.uPa * pu.s,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "mu(comp, internal energy, pressure)",
    },
    "lambda_up_func": {  # thermal conductivity
        "fname": "lambda_up",
        "units": pu.mW / pu.m / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "lambda(comp, internal energy, pressure)",
    },
    "sigma_up_func": {  # surface tension
        "fname": "sigma_up",
        "units": pu.mN / pu.m,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "sigma(comp, internal energy, pressure)",
    },
    "u_vap_up_func": {  # vapor internal energy
        "fname": "u_vap_up",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "u_vap(comp, internal energy, pressure)",
    },
    "h_vap_up_func": {  # enthalpy
        "fname": "h_vap_up",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "h_vap(comp, internal energy, pressure)",
    },
    "s_vap_up_func": {  # entropy
        "fname": "s_vap_up",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "s_vap(comp, internal energy, pressure)",
    },
    "g_vap_up_func": {  # Gibbs free energy
        "fname": "g_vap_up",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "g_vap(comp, internal energy, pressure)",
    },
    "f_vap_up_func": {  # Helmholtz free energy
        "fname": "f_vap_up",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "f_vap(comp, internal energy, pressure)",
    },
    "cv_vap_up_func": {  # constant volume heat capacity
        "fname": "cv_vap_up",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "cv_vap(comp, internal energy, pressure)",
    },
    "cp_vap_up_func": {  # constant pressure heat capacity
        "fname": "cp_vap_up",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "cp_vap(comp, internal energy, pressure)",
    },
    "w_vap_up_func": {  # speed of sound
        "fname": "w_vap_up",
        "units": pu.m / pu.s,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "w_vap(comp, internal energy, pressure)",
    },
    "v_vap_up_func": {  # specific volume
        "fname": "v_vap_up",
        "units": pu.m**3 / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "v_vap(comp, internal energy, pressure)",
    },
    "mu_vap_up_func": {  # viscosity
        "fname": "mu_vap_up",
        "units": pu.uPa * pu.s,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "mu_vap(comp, internal energy, pressure)",
    },
    "lambda_vap_up_func": {  # thermal conductivity
        "fname": "lambda_vap_up",
        "units": pu.mW / pu.m / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "lambda_vap(comp, internal energy, pressure)",
    },
    "sigma_vap_up_func": {  # surface tension
        "fname": "sigma_vap_up",
        "units": pu.mN / pu.m,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "sigma_vap(comp, internal energy, pressure)",
    },
    "u_liq_up_func": {  # liquid internal energy
        "fname": "u_liq_up",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "u_liq(comp, internal energy, pressure)",
    },
    "h_liq_up_func": {  # enthalpy
        "fname": "h_liq_up",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "h_liq(comp, internal energy, pressure)",
    },
    "s_liq_up_func": {  # entropy
        "fname": "s_liq_up",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "s_liq(comp, internal energy, pressure)",
    },
    "g_liq_up_func": {  # Gibbs free energy
        "fname": "g_liq_up",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "g_liq(comp, internal energy, pressure)",
    },
    "f_liq_up_func": {  # Helmholtz free energy
        "fname": "f_liq_up",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "f_liq(comp, internal energy, pressure)",
    },
    "cv_liq_up_func": {  # constant volume heat capacity
        "fname": "cv_liq_up",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "cv_liq(comp, internal energy, pressure)",
    },
    "cp_liq_up_func": {  # constant pressure heat capacity
        "fname": "cp_liq_up",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "cp_liq(comp, internal energy, pressure)",
    },
    "w_liq_up_func": {  # speed of sound
        "fname": "w_liq_up",
        "units": pu.m / pu.s,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "w_liq(comp, internal energy, pressure)",
    },
    "v_liq_up_func": {  # specific volume
        "fname": "v_liq_up",
        "units": pu.m**3 / pu.kg,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "v_liq(comp, internal energy, pressure)",
    },
    "mu_liq_up_func": {  # viscosity
        "fname": "mu_liq_up",
        "units": pu.uPa * pu.s,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "mu_liq(comp, internal energy, pressure)",
    },
    "lambda_liq_up_func": {  # thermal conductivity
        "fname": "lambda_liq_up",
        "units": pu.mW / pu.m / pu.K,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "lambda_liq(comp, internal energy, pressure)",
    },
    "sigma_liq_up_func": {  # surface tension
        "fname": "sigma_liq_up",
        "units": pu.mN / pu.m,
        "arg_units": [pu.dimensionless, pu.kJ / pu.kg, pu.kPa],
        "doc": "sigma_liq(comp, internal energy, pressure)",
    },
    # Functions of (T, p)
    "h_vap_tp_func": {  # enthalpy
        "fname": "h_vap_tp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "h_vap(comp, temperature, pressure)",
    },
    "u_vap_tp_func": {  # internal energy
        "fname": "u_vap_tp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "u_vap(comp, temperature, pressure)",
    },
    "s_vap_tp_func": {  # entropy
        "fname": "s_vap_tp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "s_vap(comp, temperature, pressure)",
    },
    "g_vap_tp_func": {  # Gibbs free energy
        "fname": "g_vap_tp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "g_vap(comp, temperature, pressure)",
    },
    "f_vap_tp_func": {  # Helmholtz free energy
        "fname": "f_vap_tp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "f_vap(comp, temperature, pressure)",
    },
    "cv_vap_tp_func": {  # constant volume heat capacity
        "fname": "cv_vap_tp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "cv_vap(comp, temperature, pressure)",
    },
    "cp_vap_tp_func": {  # constant pressure heat capacity
        "fname": "cp_vap_tp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "cp_vap(comp, temperature, pressure)",
    },
    "w_vap_tp_func": {  # speed of sound
        "fname": "w_vap_tp",
        "units": pu.m / pu.s,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "w_vap(comp, temperature, pressure)",
    },
    "v_vap_tp_func": {  # specific volume
        "fname": "v_vap_tp",
        "units": pu.m**3 / pu.kg,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "v_vap(comp, temperature, pressure)",
    },
    "mu_vap_tp_func": {  # viscosity
        "fname": "mu_vap_tp",
        "units": pu.uPa * pu.s,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "mu_vap(comp, temperature, pressure)",
    },
    "lambda_vap_tp_func": {  # thermal conductivity
        "fname": "lambda_vap_tp",
        "units": pu.mW / pu.m / pu.K,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "lambda_vap(comp, temperature, pressure)",
    },
    "sigma_vap_tp_func": {  # surface tension
        "fname": "sigma_vap_tp",
        "units": pu.mN / pu.m,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "sigma_vap(comp, temperature, pressure)",
    },
    "h_liq_tp_func": {  # enthalpy
        "fname": "h_liq_tp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "h_liq(comp, temperature, pressure)",
    },
    "u_liq_tp_func": {  # internal energy
        "fname": "u_liq_tp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "u_liq(comp, temperature, pressure)",
    },
    "s_liq_tp_func": {  # entropy
        "fname": "s_liq_tp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "s_liq(comp, temperature, pressure)",
    },
    "g_liq_tp_func": {  # Gibbs free energy
        "fname": "g_liq_tp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "g_liq(comp, temperature, pressure)",
    },
    "f_liq_tp_func": {  # Helmholtz free energy
        "fname": "f_liq_tp",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "f_liq(comp, temperature, pressure)",
    },
    "cv_liq_tp_func": {  # constant volume heat capacity
        "fname": "cv_liq_tp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "cv_liq(comp, temperature, pressure)",
    },
    "cp_liq_tp_func": {  # constant pressure heat capacity
        "fname": "cp_liq_tp",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "cp_liq(comp, temperature, pressure)",
    },
    "w_liq_tp_func": {  # speed of sound
        "fname": "w_liq_tp",
        "units": pu.m / pu.s,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "w_liq(comp, temperature, pressure)",
    },
    "v_liq_tp_func": {  # specific volume
        "fname": "v_liq_tp",
        "units": pu.m**3 / pu.kg,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "v_liq(comp, temperature, pressure)",
    },
    "mu_liq_tp_func": {  # viscosity
        "fname": "mu_liq_tp",
        "units": pu.uPa * pu.s,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "mu_liq(comp, temperature, pressure)",
    },
    "lambda_liq_tp_func": {  # thermal conductivity
        "fname": "lambda_liq_tp",
        "units": pu.mW / pu.m / pu.K,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "lambda_liq(comp, temperature, pressure)",
    },
    "sigma_liq_tp_func": {  # surface tension
        "fname": "sigma_liq_tp",
        "units": pu.mN / pu.m,
        "arg_units": [pu.dimensionless, pu.K, pu.kPa],
        "doc": "sigma_liq(comp, temperature, pressure)",
    },
    # Dimensionless Helmholtz energy to calculate other thermo properties
    "phi0_func": {  # ideal part
        "fname": "phi0",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
        "doc": "phi0(comp, delta, tau)",
    },
    "phi0_d_func": {  # ideal part derivative wrt delta
        "fname": "phi0_d",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
    },
    "phi0_dd_func": {  # ideal part second derivative wrt delta
        "fname": "phi0_dd",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
    },
    "phi0_t_func": {  # ideal part derivative wrt tau
        "fname": "phi0_t",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
    },
    "phi0_dt_func": {  # ideal part second derivative wrt delta and tau
        "fname": "phi0_dt",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
    },
    "phi0_tt_func": {  # ideal part second derivative wrt tau
        "fname": "phi0_tt",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
    },
    "phir_func": {  # residual part
        "fname": "phir",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
    },
    "phir_d_func": {  # residual part derivative wrt delta
        "fname": "phir_d",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
    },
    "phir_dd_func": {  # residual part second derivative wrt delta
        "fname": "phir_dd",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
    },
    "phir_t_func": {  # residual part derivative wrt tau
        "fname": "phir_t",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
    },
    "phir_dt_func": {  # residual part second derivative wrt delta and tau
        "fname": "phir_dt",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
    },
    "phir_tt_func": {  # residual part second derivative wrt tau
        "fname": "phir_tt",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.dimensionless, pu.dimensionless],
    },
    # Phase specific functions of pressure and tau
    "hvpt_func": {  # vapor enthalpy
        "fname": "hvpt",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kPa, pu.dimensionless],
        "doc": "h_v(p, tau)",
        "latex_symbol": "h_v",
    },
    "hlpt_func": {  # liquid enthalpy
        "fname": "hlpt",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kPa, pu.dimensionless],
    },
    "svpt_func": {  # vapor entropy
        "fname": "svpt",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kPa, pu.dimensionless],
    },
    "slpt_func": {  # liquid entropy
        "fname": "slpt",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kPa, pu.dimensionless],
    },
    "uvpt_func": {  # vapor internal energy
        "fname": "uvpt",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kPa, pu.dimensionless],
    },
    "ulpt_func": {  # liquid internal energy
        "fname": "ulpt",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kPa, pu.dimensionless],
    },
    "delta_liq_func": {  # liquid density
        "fname": "delta_liq",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.kPa, pu.dimensionless],
    },
    "delta_vap_func": {  # vapor density
        "fname": "delta_vap",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.kPa, pu.dimensionless],
    },
    # saturation curve as a function of tau
    "p_sat_func": {
        "fname": "p_sat",
        "units": pu.kPa,
        "arg_units": [pu.dimensionless, pu.dimensionless],
    },
    "delta_sat_v_func": {
        "fname": "delta_sat_v",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.dimensionless],
    },
    "delta_sat_l_func": {
        "fname": "delta_sat_l",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.dimensionless],
    },
    # saturation curve as a function of T
    "p_sat_t_func": {
        "fname": "p_sat_t",
        "units": pu.kPa,
        "arg_units": [pu.dimensionless, pu.K],
    },
    "h_liq_sat_t_func": {
        "fname": "h_liq_sat_t",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.K],
    },
    "h_vap_sat_t_func": {
        "fname": "h_vap_sat_t",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.K],
    },
    "s_liq_sat_t_func": {
        "fname": "s_liq_sat_t",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.K],
    },
    "s_vap_sat_t_func": {
        "fname": "s_vap_sat_t",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.K],
    },
    "u_liq_sat_t_func": {
        "fname": "u_liq_sat_t",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.K],
    },
    "u_vap_sat_t_func": {
        "fname": "u_vap_sat_t",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.K],
    },
    "v_liq_sat_t_func": {
        "fname": "v_liq_sat_t",
        "units": pu.m**3 / pu.kg,
        "arg_units": [pu.dimensionless, pu.K],
    },
    "v_vap_sat_t_func": {
        "fname": "v_vap_sat_t",
        "units": pu.m**3 / pu.kg,
        "arg_units": [pu.dimensionless, pu.K],
    },
    # saturation curve as a function of p
    "tau_sat_func": {
        "fname": "tau_sat",
        "units": pu.dimensionless,
        "arg_units": [pu.dimensionless, pu.kPa],
    },
    "t_sat_func": {
        "fname": "T_sat",
        "units": pu.K,
        "arg_units": [pu.dimensionless, pu.kPa],
    },
    "h_liq_sat_p_func": {
        "fname": "h_liq_sat_p",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kPa],
    },
    "h_vap_sat_p_func": {
        "fname": "h_vap_sat_p",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kPa],
    },
    "s_liq_sat_p_func": {
        "fname": "s_liq_sat_p",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kPa],
    },
    "s_vap_sat_p_func": {
        "fname": "s_vap_sat_p",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless, pu.kPa],
    },
    "u_liq_sat_p_func": {
        "fname": "u_liq_sat_p",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kPa],
    },
    "u_vap_sat_p_func": {
        "fname": "u_vap_sat_p",
        "units": pu.kJ / pu.kg,
        "arg_units": [pu.dimensionless, pu.kPa],
    },
    "v_liq_sat_p_func": {
        "fname": "v_liq_sat_p",
        "units": pu.m**3 / pu.kg,
        "arg_units": [pu.dimensionless, pu.kPa],
    },
    "v_vap_sat_p_func": {
        "fname": "v_vap_sat_p",
        "units": pu.m**3 / pu.kg,
        "arg_units": [pu.dimensionless, pu.kPa],
    },
    # Parameters (these functions take no arguments)
    "mw_func": {
        "fname": "mw",
        "units": pu.g / pu.mol,
        "arg_units": [pu.dimensionless],
    },
    "sgc_func": {
        "fname": "sgc",
        "units": pu.kJ / pu.kg / pu.K,
        "arg_units": [pu.dimensionless],
    },
    "t_star_func": {
        "fname": "t_star",
        "units": pu.K,
        "arg_units": [pu.dimensionless],
    },
    "rho_star_func": {
        "fname": "rho_star",
        "units": pu.kg / pu.m**3,
        "arg_units": [pu.dimensionless],
    },
    "pc_func": {
        "fname": "pc",
        "units": pu.kPa,
        "arg_units": [pu.dimensionless],
    },
    "tc_func": {
        "fname": "tc",
        "units": pu.K,
        "arg_units": [pu.dimensionless],
    },
    "rhoc_func": {
        "fname": "rhoc",
        "units": pu.kg / pu.m**3,
        "arg_units": [pu.dimensionless],
    },
    "pt_func": {
        "fname": "pt",
        "units": pu.kPa,
        "arg_units": [pu.dimensionless],
    },
    "tt_func": {
        "fname": "tt",
        "units": pu.K,
        "arg_units": [pu.dimensionless],
    },
    "rhot_l_func": {
        "fname": "rhot_l",
        "units": pu.kg / pu.m**3,
        "arg_units": [pu.dimensionless],
    },
    "rhot_v_func": {
        "fname": "rhot_v",
        "units": pu.kg / pu.m**3,
        "arg_units": [pu.dimensionless],
    },
    "pmin_func": {
        "fname": "pmin",
        "units": pu.kPa,
        "arg_units": [pu.dimensionless],
    },
    "tmin_func": {
        "fname": "tmin",
        "units": pu.K,
        "arg_units": [pu.dimensionless],
    },
    "pmax_func": {
        "fname": "pmax",
        "units": pu.kPa,
        "arg_units": [pu.dimensionless],
    },
    "tmax_func": {
        "fname": "tmax",
        "units": pu.K,
        "arg_units": [pu.dimensionless],
    },
}



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
from .helmholtz_functions import (
    StateVars,
    PhaseType,
    AmountBasis,
    HelmholtzThermoExpressions,
    HelmholtzParameterBlock,
    HelmholtzParameterBlockData,
    add_helmholtz_external_functions,
    helmholtz_available,
    helmholtz_data_dir,
)
from .helmholtz_state import (
    HelmholtzStateBlock,
    HelmholtzStateBlockData,
    HelmholtzEoSInitializer,
)
from idaes.models.properties.general_helmholtz.components.parameters import (
    get_parameter_path,
    set_parameter_path,
)
from idaes.models.properties.general_helmholtz.components.registry import (
    register_helmholtz_component,
    registered_components,
    viscosity_available,
    thermal_conductivity_available,
    surface_tension_available,
    component_registered,
    clear_component_registry,
    eos_reference,
    viscosity_reference,
    thermal_conductivity_reference,
    surface_tension_reference,
)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Commandline Utilities for Managing the IDAES Config files"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# TODO: protected access issues
# pylint: disable=protected-access

__author__ = "John Eslick"

import click

from idaes.commands import cb


@cb.command(
    name="config-write",
    help="This command has been removed.",
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
def config_write():
    click.echo("This command has been removed.")


@cb.command(
    name="config-file",
    help="This command has been removed.",
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
def config_file():
    click.echo("This command has been removed.")


@cb.command(
    name="config-set",
    help="This command has been removed.",
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
def config_set():
    click.echo("This command has been removed.")


@cb.command(
    name="config-display",
    help="This command has been removed.",
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
def config_display():
    click.echo("This command has been removed.")



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Generic Helmholtz EOS StateBlock Class
"""

__author__ = "John Eslick"

import copy

import pyomo.environ as pyo
from pyomo.common.collections import ComponentSet
from pyomo.environ import units as pyunits

from idaes.core.util.math import smooth_max
from idaes.core import declare_process_block_class
from idaes.core import (
    StateBlock,
    StateBlockData,
    MaterialBalanceType,
    EnergyBalanceType,
    MaterialFlowBasis,
)
from idaes.models.properties.general_helmholtz.helmholtz_functions import (
    add_helmholtz_external_functions,
    HelmholtzThermoExpressions,
    AmountBasis,
    PhaseType,
    StateVars,
    _data_dir,
)
from idaes.models.properties.general_helmholtz.components import (
    viscosity_available,
    thermal_conductivity_available,
    surface_tension_available,
)
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog
from idaes.core.util.initialization import fix_state_vars
from idaes.core.initialization.initializer_base import (
    InitializerBase,
    InitializationStatus,
)

_log = idaeslog.getLogger(__name__)


class HelmholtzEoSInitializer(InitializerBase):
    """
    Initializer object for Helmholtz EoS packages using external functions.

    Due to the use of external functions, Helmholtz EoS StateBlocks have no constraints,
    thus there is nothing to initialize. This Initializer replaces the general initialize
    method with a no-op.

    """

    CONFIG = InitializerBase.CONFIG()

    def initialize(
        self,
        model: pyo.Block,
        output_level=None,
    ):
        """
        Initialize method for Helmholtz EoS state blocks. This is a no-op.

        Args:
            model: model to be initialized
            output_level: (optional) output level to use during initialization run (overrides global setting).

        Returns:
            InitializationStatus.Ok
        """
        self._update_summary(model, "status", InitializationStatus.Ok)
        return self.summary[model]["status"]


class _StateBlock(StateBlock):
    """
    This class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """

    # Set default initializer
    default_initializer = HelmholtzEoSInitializer

    @staticmethod
    def _set_fixed(v, f):
        if f:
            v.fix()
        else:
            v.unfix()

    @staticmethod
    def _set_not_fixed(v, state, key, hold):
        if state is not None:
            if not v.fixed:
                try:
                    v.value = state[key]
                except KeyError:
                    pass
        if hold:
            v.fix()

    def fix_initialization_states(self):
        """
        Fixes state variables for state blocks.

        Returns:
            None
        """
        # Fix state variables
        fix_state_vars(self)

    def initialize(self, *args, **kwargs):
        flags = {}
        hold_state = kwargs.pop("hold_state", False)
        state_args = kwargs.pop("state_args", None)
        for i, v in self.items():
            pp = self[i].config.parameters.config.phase_presentation
            sv = self[i].state_vars
            ab = self[i].amount_basis
            if sv == StateVars.PH and ab == AmountBasis.MOLE:
                flags[i] = (v.flow_mol.fixed, v.enth_mol.fixed, v.pressure.fixed)
                self._set_not_fixed(v.flow_mol, state_args, "flow_mol", hold_state)
                self._set_not_fixed(v.enth_mol, state_args, "enth_mol", hold_state)
                self._set_not_fixed(v.pressure, state_args, "pressure", hold_state)
            elif sv == StateVars.PH and ab == AmountBasis.MASS:
                flags[i] = (v.flow_mass.fixed, v.enth_mass.fixed, v.pressure.fixed)
                self._set_not_fixed(v.flow_mass, state_args, "flow_mass", hold_state)
                self._set_not_fixed(v.enth_mass, state_args, "enth_mass", hold_state)
                self._set_not_fixed(v.pressure, state_args, "pressure", hold_state)
            elif sv == StateVars.PS and ab == AmountBasis.MOLE:
                flags[i] = (v.flow_mol.fixed, v.entr_mol.fixed, v.pressure.fixed)
                self._set_not_fixed(v.flow_mol, state_args, "flow_mol", hold_state)
                self._set_not_fixed(v.entr_mol, state_args, "enth_mol", hold_state)
                self._set_not_fixed(v.pressure, state_args, "pressure", hold_state)
            elif sv == StateVars.PS and ab == AmountBasis.MASS:
                flags[i] = (v.flow_mass.fixed, v.entr_mass.fixed, v.pressure.fixed)
                self._set_not_fixed(v.flow_mass, state_args, "flow_mass", hold_state)
                self._set_not_fixed(v.entr_mass, state_args, "enth_mass", hold_state)
                self._set_not_fixed(v.pressure, state_args, "pressure", hold_state)
            elif sv == StateVars.PU and ab == AmountBasis.MOLE:
                flags[i] = (
                    v.flow_mol.fixed,
                    v.energy_internal_mol.fixed,
                    v.pressure.fixed,
                )
                self._set_not_fixed(
                    v.energy_internal_mol, state_args, "energy_internal_mol", hold_state
                )
                self._set_not_fixed(
                    v.energy_internal_mol, state_args, "energy_internal_mol", hold_state
                )
                self._set_not_fixed(
                    v.energy_internal_mol, state_args, "energy_internal_mol", hold_state
                )
            elif sv == StateVars.PU and ab == AmountBasis.MASS:
                flags[i] = (
                    v.flow_mass.fixed,
                    v.energy_internal_mass.fixed,
                    v.pressure.fixed,
                )
                self._set_not_fixed(
                    v.energy_internal_mass,
                    state_args,
                    "energy_internal_mass",
                    hold_state,
                )
                self._set_not_fixed(
                    v.energy_internal_mass,
                    state_args,
                    "energy_internal_mass",
                    hold_state,
                )
                self._set_not_fixed(v.pressure, state_args, "pressure", hold_state)
            elif sv == StateVars.TPX and ab == AmountBasis.MOLE:
                # Hold the T-P-x vars
                if pp in (PhaseType.MIX, PhaseType.LG):
                    flags[i] = (
                        v.flow_mol.fixed,
                        v.temperature.fixed,
                        v.pressure.fixed,
                        v.vapor_frac.fixed,
                    )
                    self._set_not_fixed(v.flow_mol, state_args, "flow_mol", hold_state)
                    self._set_not_fixed(
                        v.temperature, state_args, "temperature", hold_state
                    )
                    self._set_not_fixed(v.pressure, state_args, "pressure", hold_state)
                    self._set_not_fixed(
                        v.vapor_frac, state_args, "vapor_frac", hold_state
                    )
                else:
                    flags[i] = (v.flow_mol.fixed, v.temperature.fixed, v.pressure.fixed)
                    self._set_not_fixed(v.flow_mol, state_args, "flow_mol", hold_state)
                    self._set_not_fixed(
                        v.temperature, state_args, "temperature", hold_state
                    )
                    self._set_not_fixed(v.pressure, state_args, "pressure", hold_state)
            elif sv == StateVars.TPX and ab == AmountBasis.MASS:
                # Hold the T-P-x vars
                if pp in (PhaseType.MIX, PhaseType.LG):
                    flags[i] = (
                        v.flow_mass.fixed,
                        v.temperature.fixed,
                        v.pressure.fixed,
                        v.vapor_frac.fixed,
                    )
                    self._set_not_fixed(
                        v.flow_mass, state_args, "flow_mass", hold_state
                    )
                    self._set_not_fixed(
                        v.temperature, state_args, "temperature", hold_state
                    )
                    self._set_not_fixed(v.pressure, state_args, "pressure", hold_state)
                    self._set_not_fixed(
                        v.vapor_frac, state_args, "vapor_frac", hold_state
                    )
                else:
                    flags[i] = (
                        v.flow_mass.fixed,
                        v.temperature.fixed,
                        v.pressure.fixed,
                    )
                    self._set_not_fixed(
                        v.flow_mass, state_args, "flow_mass", hold_state
                    )
                    self._set_not_fixed(
                        v.temperature, state_args, "temperature", hold_state
                    )
                    self._set_not_fixed(v.pressure, state_args, "pressure", hold_state)

        return flags

    def release_state(self, flags, **kwargs):
        """Set the variables back to there original fixed/free state
        after initialization.

        Args:
            flags (dict): Original variable states
        """
        for i, f in flags.items():
            pp = self[i].config.parameters.config.phase_presentation
            sv = self[i].state_vars
            ab = self[i].amount_basis
            if sv == StateVars.PH and ab == AmountBasis.MOLE:
                self._set_fixed(self[i].flow_mol, f[0])
                self._set_fixed(self[i].enth_mol, f[1])
                self._set_fixed(self[i].pressure, f[2])
            elif sv == StateVars.PH and ab == AmountBasis.MASS:
                self._set_fixed(self[i].flow_mass, f[0])
                self._set_fixed(self[i].enth_mass, f[1])
                self._set_fixed(self[i].pressure, f[2])
            elif sv == StateVars.PS and ab == AmountBasis.MOLE:
                self._set_fixed(self[i].flow_mol, f[0])
                self._set_fixed(self[i].entr_mol, f[1])
                self._set_fixed(self[i].pressure, f[2])
            elif sv == StateVars.PS and ab == AmountBasis.MASS:
                self._set_fixed(self[i].flow_mass, f[0])
                self._set_fixed(self[i].entr_mass, f[1])
                self._set_fixed(self[i].pressure, f[2])
            elif sv == StateVars.PU and ab == AmountBasis.MOLE:
                self._set_fixed(self[i].flow_mol, f[0])
                self._set_fixed(self[i].energy_internal_mol, f[1])
                self._set_fixed(self[i].pressure, f[2])
            elif sv == StateVars.PU and ab == AmountBasis.MASS:
                self._set_fixed(self[i].flow_mass, f[0])
                self._set_fixed(self[i].energy_internal_mass, f[1])
                self._set_fixed(self[i].pressure, f[2])
            elif sv == StateVars.TPX and ab == AmountBasis.MOLE:
                self._set_fixed(self[i].flow_mol, f[0])
                self._set_fixed(self[i].temperature, f[1])
                self._set_fixed(self[i].pressure, f[2])
                if pp in (PhaseType.MIX, PhaseType.LG):
                    self._set_fixed(self[i].vapor_frac, f[3])
            elif sv == StateVars.TPX and ab == AmountBasis.MASS:
                self._set_fixed(self[i].flow_mass, f[0])
                self._set_fixed(self[i].temperature, f[1])
                self._set_fixed(self[i].pressure, f[2])
                if pp in (PhaseType.MIX, PhaseType.LG):
                    self._set_fixed(self[i].vapor_frac, f[3])


@declare_process_block_class("HelmholtzStateBlock", block_class=_StateBlock)
class HelmholtzStateBlockData(StateBlockData):
    """
    This is a base class for Helmholtz equations of state using IDAES standard
    Helmholtz EOS external functions written in C++.
    """

    def _state_vars(self):
        """Create the state variables"""
        params = self.config.parameters
        cmp = params.pure_component
        phase_set = params.config.phase_presentation

        # Check for inconsistent phase presentation and phase equilibrium
        if phase_set == PhaseType.MIX and self.config.has_phase_equilibrium:
            _log.warning(
                "Helmholtz EoS packages using Mixed phase representation ignore the "
                "'has_phase_equilibrium' configuration argument. However, setting "
                "this to True can result in errors when constructing material balances "
                "due to only having a single phase (thus phase transfer terms cannot "
                "be constructed)."
            )

        # Add flow variable
        if self.amount_basis == AmountBasis.MOLE:
            self.flow_mol = pyo.Var(
                initialize=1, doc="Total mole flow", units=pyunits.mol / pyunits.s
            )
            self.flow_mass = pyo.Expression(
                expr=self.mw * self.flow_mol, doc="Total mass flow"
            )
        else:
            self.flow_mass = pyo.Var(
                initialize=1, doc="Total mass flow", units=pyunits.kg / pyunits.s
            )
            self.flow_mol = pyo.Expression(
                expr=self.flow_mass / self.mw, doc="Total mole flow"
            )
        # All supported state variable sets include pressure
        self.pressure = pyo.Var(
            domain=pyo.PositiveReals,
            initialize=params.default_pressure_value,
            doc="Pressure",
            bounds=params.default_pressure_bounds,
            units=pyunits.Pa,
        )
        self.p_kPa = pyo.Expression(expr=self.pressure * params.uc["Pa to kPa"])
        # If it's single phase use provide fixed expressions for vapor frac
        if phase_set == PhaseType.L:
            self.vapor_frac = pyo.Expression(
                expr=0.0, doc="Vapor mole fraction (mol vapor/mol total)"
            )
        elif phase_set == PhaseType.G:
            self.vapor_frac = pyo.Expression(
                expr=1.0,
                doc="Vapor mole fraction (mol vapor/mol total)",
            )
        # Add other state var
        if self.state_vars == StateVars.PH and self.amount_basis == AmountBasis.MOLE:
            self.enth_mol = pyo.Var(
                initialize=params.default_enthalpy_mol_value,
                doc="Total molar enthalpy",
                bounds=params.default_enthalpy_mol_bounds,
                units=pyunits.J / pyunits.mol,
            )
            self.h_kJ_per_kg = pyo.Expression(
                expr=self.enth_mol * params.uc["J/mol to kJ/kg"]
            )
            self.temperature = pyo.Expression(
                expr=self.t_hp_func(cmp, self.h_kJ_per_kg, self.p_kPa, _data_dir),
                doc="Temperature",
            )
            if phase_set == PhaseType.MIX or phase_set == PhaseType.LG:
                self.vapor_frac = pyo.Expression(
                    expr=self.vf_hp_func(cmp, self.h_kJ_per_kg, self.p_kPa, _data_dir),
                    doc="Vapor mole fraction (mol vapor/mol total)",
                )
            self._state_vars_dict = {
                "flow_mol": self.flow_mol,
                "enth_mol": self.enth_mol,
                "pressure": self.pressure,
            }
            self.extensive_set = ComponentSet((self.flow_mol,))
            self.intensive_set = ComponentSet((self.enth_mol, self.pressure))
        elif self.state_vars == StateVars.PH and self.amount_basis == AmountBasis.MASS:
            self.enth_mass = pyo.Var(
                initialize=params.default_enthalpy_mass_value,
                doc="Total enthalpy per mass",
                bounds=params.default_enthalpy_mass_bounds,
                units=pyunits.J / pyunits.kg,
            )
            self.h_kJ_per_kg = pyo.Expression(
                expr=self.enth_mass * params.uc["J/kg to kJ/kg"]
            )
            self.temperature = pyo.Expression(
                expr=self.t_hp_func(cmp, self.h_kJ_per_kg, self.p_kPa, _data_dir),
                doc="Temperature",
            )
            if phase_set == PhaseType.MIX or phase_set == PhaseType.LG:
                self.vapor_frac = pyo.Expression(
                    expr=self.vf_hp_func(cmp, self.h_kJ_per_kg, self.p_kPa, _data_dir),
                    doc="Vapor mole fraction (mol vapor/mol total)",
                )
            self._state_vars_dict = {
                "flow_mass": self.flow_mass,
                "enth_mass": self.enth_mass,
                "pressure": self.pressure,
            }
            self.extensive_set = ComponentSet((self.flow_mass,))
            self.intensive_set = ComponentSet((self.enth_mass, self.pressure))
        elif self.state_vars == StateVars.PS and self.amount_basis == AmountBasis.MOLE:
            self.entr_mol = pyo.Var(
                initialize=params.default_entropy_mol_value,
                doc="Total molar entropy",
                bounds=params.default_entropy_mol_bounds,
                units=pyunits.J / pyunits.mol / pyunits.K,
            )
            self.s_kJ_per_kgK = pyo.Expression(
                expr=self.entr_mol * params.uc["J/mol/K to kJ/kg/K"]
            )
            self.temperature = pyo.Expression(
                expr=self.t_sp_func(cmp, self.s_kJ_per_kgK, self.p_kPa, _data_dir),
                doc="Temperature",
            )
            if phase_set == PhaseType.MIX or phase_set == PhaseType.LG:
                self.vapor_frac = pyo.Expression(
                    expr=self.vf_sp_func(cmp, self.s_kJ_per_kgK, self.p_kPa, _data_dir),
                    doc="Vapor mole fraction (mol vapor/mol total)",
                )
            self._state_vars_dict = {
                "flow_mol": self.flow_mol,
                "entr_mol": self.entr_mol,
                "pressure": self.pressure,
            }
            self.extensive_set = ComponentSet((self.flow_mol,))
            self.intensive_set = ComponentSet((self.entr_mol, self.pressure))
        elif self.state_vars == StateVars.PS and self.amount_basis == AmountBasis.MASS:
            self.entr_mass = pyo.Var(
                initialize=params.default_entropy_mass_value,
                doc="Total entropy per mass",
                bounds=params.default_entropy_mass_bounds,
                units=pyunits.J / pyunits.kg / pyunits.K,
            )
            self.s_kJ_per_kgK = pyo.Expression(
                expr=self.entr_mass * params.uc["J/kg/K to kJ/kg/K"]
            )
            self.temperature = pyo.Expression(
                expr=self.t_sp_func(cmp, self.s_kJ_per_kgK, self.p_kPa, _data_dir),
                doc="Temperature",
            )
            if phase_set == PhaseType.MIX or phase_set == PhaseType.LG:
                self.vapor_frac = pyo.Expression(
                    expr=self.vf_sp_func(cmp, self.s_kJ_per_kgK, self.p_kPa, _data_dir),
                    doc="Vapor mole fraction (mol vapor/mol total)",
                )
            self._state_vars_dict = {
                "flow_mass": self.flow_mass,
                "entr_mass": self.entr_mass,
                "pressure": self.pressure,
            }
            self.extensive_set = ComponentSet((self.flow_mass,))
            self.intensive_set = ComponentSet((self.entr_mass, self.pressure))
        elif self.state_vars == StateVars.PU and self.amount_basis == AmountBasis.MOLE:
            self.energy_internal_mol = pyo.Var(
                initialize=params.default_energy_internal_mol_value,
                doc="Total molar internal energy",
                bounds=params.default_energy_internal_mol_bounds,
                units=pyunits.J / pyunits.mol,
            )
            self.u_kJ_per_kg = pyo.Expression(
                expr=self.energy_internal_mol * params.uc["J/mol to kJ/kg"]
            )
            self.temperature = pyo.Expression(
                expr=self.t_up_func(cmp, self.u_kJ_per_kg, self.p_kPa, _data_dir),
                doc="Temperature",
            )
            if phase_set == PhaseType.MIX or phase_set == PhaseType.LG:
                self.vapor_frac = pyo.Expression(
                    expr=self.vf_up_func(cmp, self.u_kJ_per_kg, self.p_kPa, _data_dir),
                    doc="Vapor mole fraction (mol vapor/mol total)",
                )
            self._state_vars_dict = {
                "flow_mol": self.flow_mol,
                "energy_internal_mol": self.energy_internal_mol,
                "pressure": self.pressure,
            }
            self.extensive_set = ComponentSet((self.flow_mol,))
            self.intensive_set = ComponentSet((self.energy_internal_mol, self.pressure))
        elif self.state_vars == StateVars.PU and self.amount_basis == AmountBasis.MASS:
            self.energy_internal_mass = pyo.Var(
                initialize=params.default_energy_internal_mass_value,
                doc="Total internal energy per mass",
                bounds=params.default_energy_internal_mass_bounds,
                units=pyunits.J / pyunits.kg,
            )
            self.u_kJ_per_kg = pyo.Expression(
                expr=self.energy_internal_mass * params.uc["J/kg to kJ/kg"]
            )
            self.temperature = pyo.Expression(
                expr=self.t_up_func(cmp, self.u_kJ_per_kg, self.p_kPa, _data_dir),
                doc="Temperature",
            )
            if phase_set == PhaseType.MIX or phase_set == PhaseType.LG:
                self.vapor_frac = pyo.Expression(
                    expr=self.vf_up_func(cmp, self.u_kJ_per_kg, self.p_kPa, _data_dir),
                    doc="Vapor mole fraction (mol vapor/mol total)",
                )
            self._state_vars_dict = {
                "flow_mass": self.flow_mass,
                "energy_internal_mass": self.energy_internal_mass,
                "pressure": self.pressure,
            }
            self.extensive_set = ComponentSet((self.flow_mass,))
            self.intensive_set = ComponentSet(
                (self.energy_internal_mass, self.pressure)
            )
        if self.state_vars == StateVars.TPX:
            self.temperature = pyo.Var(
                domain=pyo.PositiveReals,
                initialize=params.default_temperature_value,
                doc="Temperature",
                bounds=params.default_temperature_bounds,
                units=pyunits.K,
            )
            if phase_set == PhaseType.MIX or phase_set == PhaseType.LG:
                self.vapor_frac = pyo.Var(
                    initialize=0.0,
                    units=pyunits.dimensionless,
                    doc="Vapor fraction",
                    # No bounds here, since it is often (usually) on it's bound
                    # and that's not the best for IPOPT
                )
                self.intensive_set = ComponentSet(
                    (self.temperature, self.pressure, self.vapor_frac)
                )
                self._state_vars_dict = {
                    "temperature": self.temperature,
                    "pressure": self.pressure,
                    "vapor_frac": self.vapor_frac,
                }
            else:
                self.intensive_set = ComponentSet((self.temperature, self.pressure))
                self._state_vars_dict = {
                    "temperature": self.temperature,
                    "pressure": self.pressure,
                }
            if self.amount_basis == AmountBasis.MOLE:
                self.extensive_set = ComponentSet((self.flow_mol,))
                self._state_vars_dict["flow_mol"] = self.flow_mol
            else:
                self.extensive_set = ComponentSet((self.flow_mass,))
                self._state_vars_dict["flow_mass"] = self.flow_mass

    def _tpx_phase_eq(self):
        params = self.config.parameters
        eps_pu = params.smoothing_pressure_under
        eps_po = params.smoothing_pressure_over
        priv_plist = params.private_phase_list
        # Convert pressures to kPa for external functions and nicer scaling in
        # the complementarity-type constraints.
        vf = self.vapor_frac
        # Terms for determining if you are above, below, or at the Psat
        self.pressure_under_sat = pyo.Expression(
            expr=smooth_max(0, self.pressure_sat - self.pressure, eps_pu),
            doc="pressure above Psat, 0 if liquid exists",
        )
        self.pressure_over_sat = pyo.Expression(
            expr=smooth_max(0, self.pressure - self.pressure_sat, eps_po),
            doc="pressure below Psat, 0 if vapor exists",
        )

        if not self.config.defined_state:
            self.eq_complementarity = pyo.Constraint(
                expr=0
                == (
                    vf * self.pressure_over_sat / 1000.0
                    - (1 - vf) * self.pressure_under_sat / 1000.0
                )
            )

        # Calculate liquid and vapor density.  If the phase doesn't exist,
        # density will be calculated at the saturation or critical pressure
        def rule_pressure_phase(b, p):
            if p == "Liq":
                return self.pressure + self.pressure_under_sat
            else:
                return self.pressure - self.pressure_over_sat

        self.pressure_phase = pyo.Expression(priv_plist, rule=rule_pressure_phase)

        # Constraint that can be activated to enforce that you are in the two-phase region
        self.eq_sat = pyo.Constraint(
            expr=self.pressure * 1e-6 == self.pressure_sat * 1e-6
        )
        self.eq_sat.deactivate()

    def build(self, *args):
        """
        Callable method for Block construction
        """
        # Call base class build
        super().build(*args)
        # Short path to the parameter block
        params = self.config.parameters
        # Check if the library is available, and add external functions.
        add_helmholtz_external_functions(self)
        cmp = params.pure_component
        # Which state vars to use
        self.state_vars = params.state_vars
        # Mass or Mole basis
        self.amount_basis = params.config.amount_basis
        # Private phase list
        phlist = params.private_phase_list
        # Public phase list
        pub_phlist = params.phase_list
        component_list = params.component_list
        self.phase_equilibrium_list = params.phase_equilibrium_list

        # Add component mole fraction for standardization
        def mole_frac_comp_rule(b, i):
            return 1.0

        self.mole_frac_comp = pyo.Expression(component_list, rule=mole_frac_comp_rule)
        # Expressions that link to some parameters in the param block, which
        # are commonly needed, this lets you get the parameters with scale
        # factors directly from the state block
        self.temperature_crit = pyo.Expression(
            expr=params.temperature_crit, doc="critical temperature"
        )
        self.temperature_star = pyo.Expression(
            expr=params.temperature_star, doc="temperature for tau calculation"
        )
        self.pressure_crit = pyo.Expression(
            expr=params.pressure_crit, doc="critical pressure"
        )
        self.dens_mass_crit = pyo.Expression(
            expr=params.dens_mass_crit, doc="critical mass density"
        )
        self.dens_mass_star = pyo.Expression(
            expr=params.dens_mass_star, doc="mass density for delta calculation"
        )
        self.dens_mol_crit = pyo.Expression(
            expr=params.dens_mass_crit / params.mw, doc="critical mole density"
        )
        self.dens_mol_star = pyo.Expression(
            expr=params.dens_mass_star / params.mw,
            doc="mole density for delta calculation",
        )
        self.mw = pyo.Expression(expr=params.mw, doc="molecular weight")
        # create the appropriate state variables and expressions for anything
        # that could be a state variable (H, S, U, P, T, X) on a mass and mole
        # basis if using TPx as state variables it also adds the complementarity
        # constraints. Beyond this everything else is common.
        self._state_vars()

        # P is pressure in kPa for external function calls
        T = self.temperature
        vf = self.vapor_frac
        # Saturation temperature expression
        self.temperature_sat = pyo.Expression(
            expr=self.t_sat_func(cmp, self.p_kPa, _data_dir),
            doc="Saturation temperature",
        )
        # Saturation pressure
        self.pressure_sat = pyo.Expression(
            expr=self.p_sat_t_func(cmp, T, _data_dir) * params.uc["kPa to Pa"],
            doc="Saturation pressure",
        )
        # Add the complementarity constraint for phase equilibrium with TPx.
        if self.state_vars == StateVars.TPX and len(phlist) > 1:
            self._tpx_phase_eq()

        # to make writing the remaining expressions simpler create a state var dict
        if self.state_vars == StateVars.PH:
            sv_dict = {"h": self.h_kJ_per_kg, "p": self.p_kPa}
        elif self.state_vars == StateVars.PS:
            sv_dict = {"s": self.s_kJ_per_kgK, "p": self.p_kPa}
        elif self.state_vars == StateVars.PU:
            sv_dict = {"u": self.u_kJ_per_kg, "p": self.p_kPa}
        elif self.state_vars == StateVars.TPX:
            sv_dict = {
                "T": self.temperature,
                "p": self.p_kPa,
                "x": self.vapor_frac,
            }
        sv_dict_liq = copy.copy(sv_dict)
        sv_dict_vap = copy.copy(sv_dict)
        if self.state_vars == StateVars.TPX and len(phlist) > 1:
            self.p_kPa_liq = pyo.Expression(
                expr=self.pressure_phase["Liq"] * params.uc["Pa to kPa"]
            )
            self.p_kPa_vap = pyo.Expression(
                expr=self.pressure_phase["Vap"] * params.uc["Pa to kPa"]
            )
            sv_dict_liq["p"] = self.p_kPa_liq
            sv_dict_vap["p"] = self.p_kPa_vap

        self.expression_writer = HelmholtzThermoExpressions(self, params)

        # Saturated Enthalpy molar
        def rule_enth_mol_sat_phase(b, p):
            if p == "Liq":
                return self.expression_writer.h_liq_sat(
                    p=b.p_kPa, result_basis=AmountBasis.MOLE, convert_args=False
                )
            else:
                return self.expression_writer.h_vap_sat(
                    p=b.p_kPa, result_basis=AmountBasis.MOLE, convert_args=False
                )

        self.enth_mol_sat_phase = pyo.Expression(
            phlist,
            rule=rule_enth_mol_sat_phase,
            doc="Saturated enthalpy of the phases at pressure",
        )

        # Saturated Enthalpy mass
        def rule_enth_mass_sat_phase(b, p):
            if p == "Liq":
                return self.expression_writer.h_liq_sat(
                    p=b.p_kPa, result_basis=AmountBasis.MASS, convert_args=False
                )
            else:
                return self.expression_writer.h_vap_sat(
                    p=b.p_kPa, result_basis=AmountBasis.MASS, convert_args=False
                )

        self.enth_mass_sat_phase = pyo.Expression(
            phlist,
            rule=rule_enth_mass_sat_phase,
            doc="Saturated enthalpy of the phases at pressure",
        )

        # Saturated Entropy molar
        def rule_entr_mol_sat_phase(b, p):
            if p == "Liq":
                return self.expression_writer.s_liq_sat(
                    p=b.p_kPa, result_basis=AmountBasis.MOLE, convert_args=False
                )
            else:
                return self.expression_writer.s_vap_sat(
                    p=b.p_kPa, result_basis=AmountBasis.MOLE, convert_args=False
                )

        self.entr_mol_sat_phase = pyo.Expression(
            phlist,
            rule=rule_entr_mol_sat_phase,
            doc="Saturated entropy of the phases at pressure",
        )

        # Saturated Entropy mass
        def rule_entr_mass_sat_phase(b, p):
            if p == "Liq":
                return self.expression_writer.s_liq_sat(
                    p=b.p_kPa, result_basis=AmountBasis.MASS, convert_args=False
                )
            else:
                return self.expression_writer.s_vap_sat(
                    p=b.p_kPa, result_basis=AmountBasis.MASS, convert_args=False
                )

        self.entr_mass_sat_phase = pyo.Expression(
            phlist,
            rule=rule_entr_mass_sat_phase,
            doc="Saturated entropy of the phases at pressure",
        )

        # Saturated internal energy molar
        def rule_energy_internal_mol_sat_phase(b, p):
            if p == "Liq":
                return self.expression_writer.u_liq_sat(
                    p=b.p_kPa, result_basis=AmountBasis.MOLE, convert_args=False
                )
            else:
                return self.expression_writer.u_vap_sat(
                    p=b.p_kPa, result_basis=AmountBasis.MOLE, convert_args=False
                )

        self.energy_internal_mol_sat_phase = pyo.Expression(
            phlist,
            rule=rule_energy_internal_mol_sat_phase,
            doc="Saturated internal energy of the phases at pressure",
        )

        # Saturated internal energy mass
        def rule_energy_internal_mass_sat_phase(b, p):
            if p == "Liq":
                return self.expression_writer.u_liq_sat(
                    p=b.p_kPa, result_basis=AmountBasis.MASS, convert_args=False
                )
            else:
                return self.expression_writer.u_vap_sat(
                    p=b.p_kPa, result_basis=AmountBasis.MASS, convert_args=False
                )

        self.energy_internal_mass_sat_phase = pyo.Expression(
            phlist,
            rule=rule_energy_internal_mass_sat_phase,
            doc="Saturated internal energy of the phases at pressure",
        )

        # Saturated molar volume
        def rule_volume_mol_sat_phase(b, p):
            if p == "Liq":
                return self.expression_writer.v_liq_sat(
                    p=self.p_kPa, result_basis=AmountBasis.MOLE, convert_args=False
                )
            else:
                return self.expression_writer.v_vap_sat(
                    p=self.p_kPa, result_basis=AmountBasis.MOLE, convert_args=False
                )

        self.volume_mol_sat_phase = pyo.Expression(
            phlist,
            rule=rule_volume_mol_sat_phase,
            doc="Saturated molar volume of the phases at pressure",
        )

        # Saturated specific volume
        def rule_volume_mass_sat_phase(b, p):
            if p == "Liq":
                return self.expression_writer.v_liq_sat(
                    p=self.p_kPa, result_basis=AmountBasis.MASS, convert_args=False
                )
            else:
                return self.expression_writer.v_vap_sat(
                    p=self.p_kPa, result_basis=AmountBasis.MASS, convert_args=False
                )

        self.volume_mass_sat_phase = pyo.Expression(
            phlist,
            rule=rule_volume_mass_sat_phase,
            doc="Saturated specific volume of the phases at pressure",
        )

        # delta h vap at P
        self.dh_vap_mol = pyo.Expression(
            expr=(
                self.expression_writer.h_vap_sat(
                    p=self.p_kPa, result_basis=AmountBasis.MOLE, convert_args=False
                )
                - self.expression_writer.h_liq_sat(
                    p=self.p_kPa, result_basis=AmountBasis.MOLE, convert_args=False
                )
            ),
            doc="Enthalpy of vaporization at pressure and saturation temperature",
        )

        # delta s vap at P
        self.ds_vap_mol = pyo.Expression(
            expr=(
                self.expression_writer.s_vap_sat(
                    p=self.p_kPa, result_basis=AmountBasis.MOLE, convert_args=False
                )
                - self.expression_writer.s_liq_sat(
                    p=self.p_kPa, result_basis=AmountBasis.MOLE, convert_args=False
                )
            ),
            doc="Entropy of vaporization at pressure and saturation temperature",
        )

        # delta u vap at P
        self.du_vap_mol = pyo.Expression(
            expr=(
                self.expression_writer.u_vap_sat(
                    p=self.p_kPa, result_basis=AmountBasis.MOLE, convert_args=False
                )
                - self.expression_writer.u_liq_sat(
                    p=self.p_kPa, result_basis=AmountBasis.MOLE, convert_args=False
                )
            ),
            doc="Internal energy of vaporization at pressure and saturation temperature",
        )

        # delta h vap at P
        self.dh_vap_mass = pyo.Expression(
            expr=(
                self.expression_writer.h_vap_sat(
                    p=self.p_kPa, result_basis=AmountBasis.MASS, convert_args=False
                )
                - self.expression_writer.h_liq_sat(
                    p=self.p_kPa, result_basis=AmountBasis.MASS, convert_args=False
                )
            ),
            doc="Enthalpy of vaporization at pressure and saturation temperature",
        )

        # delta s vap at P
        self.ds_vap_mass = pyo.Expression(
            expr=(
                self.expression_writer.s_vap_sat(
                    p=self.p_kPa, result_basis=AmountBasis.MASS, convert_args=False
                )
                - self.expression_writer.s_liq_sat(
                    p=self.p_kPa, result_basis=AmountBasis.MASS, convert_args=False
                )
            ),
            doc="Entropy of vaporization at pressure and saturation temperature",
        )

        # delta u vap at P
        self.du_vap_mass = pyo.Expression(
            expr=(
                self.expression_writer.u_vap_sat(
                    p=self.p_kPa, result_basis=AmountBasis.MASS, convert_args=False
                )
                - self.expression_writer.u_liq_sat(
                    p=self.p_kPa, result_basis=AmountBasis.MASS, convert_args=False
                )
            ),
            doc="Internal energy of vaporization at pressure and saturation temperature",
        )

        # Phase fraction
        def rule_phase_frac(b, p):
            if p == "Vap":
                return vf
            elif p == "Liq":
                return 1.0 - vf

        self.phase_frac = pyo.Expression(
            phlist, rule=rule_phase_frac, doc="Phase fraction"
        )

        def rule_energy_internal_mol_phase(b, p):
            if p == "Liq":
                return self.expression_writer.u_liq(
                    **sv_dict_liq, result_basis=AmountBasis.MOLE, convert_args=False
                )
            else:
                return self.expression_writer.u_vap(
                    **sv_dict_vap, result_basis=AmountBasis.MOLE, convert_args=False
                )

        self.energy_internal_mol_phase = pyo.Expression(
            phlist,
            rule=rule_energy_internal_mol_phase,
            doc="Phase internal energy",
        )

        def rule_energy_internal_mass_phase(b, p):
            if p == "Liq":
                return self.expression_writer.u_liq(
                    **sv_dict_liq, result_basis=AmountBasis.MASS, convert_args=False
                )
            else:
                return self.expression_writer.u_vap(
                    **sv_dict_vap, result_basis=AmountBasis.MASS, convert_args=False
                )

        self.energy_internal_mass_phase = pyo.Expression(
            phlist,
            rule=rule_energy_internal_mass_phase,
            doc="Phase internal energy",
        )

        # Phase Enthalpy
        def rule_enth_mol_phase(b, p):
            if p == "Liq":
                return self.expression_writer.h_liq(
                    **sv_dict_liq, result_basis=AmountBasis.MOLE, convert_args=False
                )
            else:
                return self.expression_writer.h_vap(
                    **sv_dict_vap, result_basis=AmountBasis.MOLE, convert_args=False
                )

        self.enth_mol_phase = pyo.Expression(
            phlist,
            rule=rule_enth_mol_phase,
            doc="Phase enthalpy",
        )

        def rule_enth_mass_phase(b, p):
            if p == "Liq":
                return self.expression_writer.h_liq(
                    **sv_dict_liq, result_basis=AmountBasis.MASS, convert_args=False
                )
            else:
                return self.expression_writer.h_vap(
                    **sv_dict_vap, result_basis=AmountBasis.MASS, convert_args=False
                )

        self.enth_mass_phase = pyo.Expression(
            phlist,
            rule=rule_enth_mass_phase,
            doc="Phase enthalpy",
        )

        # Phase Entropy
        def rule_entr_mol_phase(b, p):
            if p == "Liq":
                return self.expression_writer.s_liq(
                    **sv_dict_liq, result_basis=AmountBasis.MOLE, convert_args=False
                )
            else:
                return self.expression_writer.s_vap(
                    **sv_dict_vap, result_basis=AmountBasis.MOLE, convert_args=False
                )

        self.entr_mol_phase = pyo.Expression(
            phlist,
            rule=rule_entr_mol_phase,
            doc="Phase entropy",
        )

        def rule_entr_mass_phase(b, p):
            if p == "Liq":
                return self.expression_writer.s_liq(
                    **sv_dict_liq, result_basis=AmountBasis.MASS, convert_args=False
                )
            else:
                return self.expression_writer.s_vap(
                    **sv_dict_vap, result_basis=AmountBasis.MASS, convert_args=False
                )

        self.entr_mass_phase = pyo.Expression(
            phlist,
            rule=rule_entr_mass_phase,
            doc="Phase entropy",
        )

        # Phase constant pressure heat capacity, cp
        def rule_cp_mol_phase(b, p):
            if p == "Liq":
                return self.expression_writer.cp_liq(
                    **sv_dict_liq, result_basis=AmountBasis.MOLE, convert_args=False
                )
            else:
                return self.expression_writer.cp_vap(
                    **sv_dict_vap, result_basis=AmountBasis.MOLE, convert_args=False
                )

        self.cp_mol_phase = pyo.Expression(
            phlist,
            rule=rule_cp_mol_phase,
            doc="Phase isobaric heat capacity",
        )

        def rule_cp_mass_phase(b, p):
            if p == "Liq":
                return self.expression_writer.cp_liq(
                    **sv_dict_liq, result_basis=AmountBasis.MASS, convert_args=False
                )
            else:
                return self.expression_writer.cp_vap(
                    **sv_dict_vap, result_basis=AmountBasis.MASS, convert_args=False
                )

        self.cp_mass_phase = pyo.Expression(
            phlist,
            rule=rule_cp_mass_phase,
            doc="Phase isobaric heat capacity",
        )

        # Phase constant volume heat capacity, cv
        def rule_cv_mol_phase(b, p):
            if p == "Liq":
                return self.expression_writer.cv_liq(
                    **sv_dict_liq, result_basis=AmountBasis.MOLE, convert_args=False
                )
            else:
                return self.expression_writer.cv_vap(
                    **sv_dict_vap, result_basis=AmountBasis.MOLE, convert_args=False
                )

        self.cv_mol_phase = pyo.Expression(
            phlist,
            rule=rule_cv_mol_phase,
            doc="Phase isochoric heat capacity",
        )

        def rule_cv_mass_phase(b, p):
            if p == "Liq":
                return self.expression_writer.cv_liq(
                    **sv_dict_liq, result_basis=AmountBasis.MASS, convert_args=False
                )
            else:
                return self.expression_writer.cv_vap(
                    **sv_dict_vap, result_basis=AmountBasis.MASS, convert_args=False
                )

        self.cv_mass_phase = pyo.Expression(
            phlist,
            rule=rule_cv_mass_phase,
            doc="Phase isochoric heat capacity",
        )

        # Phase speed of sound
        def rule_speed_sound_phase(b, p):
            if p == "Liq":
                return self.expression_writer.w_liq(**sv_dict_liq, convert_args=False)
            else:
                return self.expression_writer.w_vap(**sv_dict_liq, convert_args=False)

        self.speed_sound_phase = pyo.Expression(
            phlist,
            rule=rule_speed_sound_phase,
            doc="Phase speed of sound or saturated if phase doesn't exist",
        )

        # molar volume
        def rule_vol_mol_phase(b, p):
            if p == "Liq":
                return self.expression_writer.v_liq(
                    **sv_dict_liq, result_basis=AmountBasis.MOLE, convert_args=False
                )
            else:
                return self.expression_writer.v_vap(
                    **sv_dict_vap, result_basis=AmountBasis.MOLE, convert_args=False
                )

        self.vol_mol_phase = pyo.Expression(
            phlist,
            rule=rule_vol_mol_phase,
            doc="Molar volume of phase",
        )

        # specific volume
        def rule_vol_mass_phase(b, p):
            if p == "Liq":
                return self.expression_writer.v_liq(
                    **sv_dict_liq, result_basis=AmountBasis.MASS, convert_args=False
                )
            else:
                return self.expression_writer.v_vap(
                    **sv_dict_vap, result_basis=AmountBasis.MASS, convert_args=False
                )

        self.vol_mass_phase = pyo.Expression(
            phlist,
            rule=rule_vol_mass_phase,
            doc="Specific volume of phase",
        )

        # molar density
        def rule_dens_mol_phase(b, p):
            if p == "Liq":
                return 1.0 / self.expression_writer.v_liq(
                    **sv_dict_liq, result_basis=AmountBasis.MOLE, convert_args=False
                )
            else:
                return 1.0 / self.expression_writer.v_vap(
                    **sv_dict_vap, result_basis=AmountBasis.MOLE, convert_args=False
                )

        self.dens_mol_phase = pyo.Expression(
            phlist,
            rule=rule_dens_mol_phase,
            doc="Mole density of phase",
        )

        # specific density
        def rule_dens_mass_phase(b, p):
            if p == "Liq":
                return 1.0 / self.expression_writer.v_liq(
                    **sv_dict_liq, result_basis=AmountBasis.MASS, convert_args=False
                )
            else:
                return 1.0 / self.expression_writer.v_vap(
                    **sv_dict_vap, result_basis=AmountBasis.MASS, convert_args=False
                )

        self.dens_mass_phase = pyo.Expression(
            phlist,
            rule=rule_dens_mass_phase,
            doc="Mass density of phase",
        )

        # Component flow (for units that need it)
        def component_flow_mol(b, i):
            return self.flow_mol

        self.flow_mol_comp = pyo.Expression(
            component_list,
            rule=component_flow_mol,
            doc="Total mole flow (both phases) of component",
        )

        def component_flow_mass(b, i):
            return self.flow_mass

        self.flow_mass_comp = pyo.Expression(
            component_list,
            rule=component_flow_mass,
            doc="Total mass flow (both phases) of component",
        )

        #
        # Total (mixed phase) properties
        #

        # Enthalpy and pressure state variables two-phase properties
        if self.state_vars == StateVars.PH:
            if self.amount_basis == AmountBasis.MOLE:
                self.enth_mass = pyo.Expression(
                    expr=self.enth_mol * params.uc["J/mol to J/kg"]
                )
            else:
                self.enth_mol = pyo.Expression(
                    expr=self.enth_mass * params.uc["J/kg to J/mol"]
                )
            self.entr_mass = pyo.Expression(
                expr=self.s_hp_func(cmp, self.h_kJ_per_kg, self.p_kPa, _data_dir)
                * params.uc["kJ/kg/K to J/kg/K"]
            )
            self.entr_mol = pyo.Expression(
                expr=self.s_hp_func(cmp, self.h_kJ_per_kg, self.p_kPa, _data_dir)
                * params.uc["kJ/kg/K to J/mol/K"]
            )
            self.energy_internal_mass = pyo.Expression(
                expr=self.u_hp_func(cmp, self.h_kJ_per_kg, self.p_kPa, _data_dir)
                * params.uc["kJ/kg to J/kg"]
            )
            self.energy_internal_mol = pyo.Expression(
                expr=self.u_hp_func(cmp, self.h_kJ_per_kg, self.p_kPa, _data_dir)
                * params.uc["kJ/kg to J/mol"]
            )
            self.cp_mass = pyo.Expression(
                expr=self.cp_hp_func(cmp, self.h_kJ_per_kg, self.p_kPa, _data_dir)
                * params.uc["kJ/kg/K to J/kg/K"]
            )
            self.cp_mol = pyo.Expression(
                expr=self.cp_hp_func(cmp, self.h_kJ_per_kg, self.p_kPa, _data_dir)
                * params.uc["kJ/kg/K to J/mol/K"]
            )
            self.cv_mass = pyo.Expression(
                expr=self.cv_hp_func(cmp, self.h_kJ_per_kg, self.p_kPa, _data_dir)
                * params.uc["kJ/kg/K to J/kg/K"]
            )
            self.cv_mol = pyo.Expression(
                expr=self.cv_hp_func(cmp, self.h_kJ_per_kg, self.p_kPa, _data_dir)
                * params.uc["kJ/kg/K to J/mol/K"]
            )
            self.dens_mass = pyo.Expression(
                expr=1.0 / self.v_hp_func(cmp, self.h_kJ_per_kg, self.p_kPa, _data_dir)
            )
            self.dens_mol = pyo.Expression(
                expr=params.uc["kg/m3 to mol/m3"]
                / self.v_hp_func(cmp, self.h_kJ_per_kg, self.p_kPa, _data_dir)
            )
        # Entropy is a state variable
        elif self.state_vars == StateVars.PS:
            if self.amount_basis == AmountBasis.MOLE:
                self.entr_mass = pyo.Expression(
                    expr=self.entr_mol * params.uc["J/mol/K to J/kg/K"]
                )
            else:
                self.enth_mol = pyo.Expression(
                    expr=self.entr_mass * params.uc["J/kg/K to J/mol/K"]
                )
            self.enth_mass = pyo.Expression(
                expr=self.h_sp_func(cmp, self.s_kJ_per_kgK, self.p_kPa, _data_dir)
                * params.uc["kJ/kg to J/kg"]
            )
            self.enth_mol = pyo.Expression(
                expr=self.h_sp_func(cmp, self.s_kJ_per_kgK, self.p_kPa, _data_dir)
                * params.uc["kJ/kg to J/mol"]
            )
            self.energy_internal_mass = pyo.Expression(
                expr=self.u_sp_func(cmp, self.s_kJ_per_kgK, self.p_kPa, _data_dir)
                * params.uc["kJ/kg to J/kg"]
            )
            self.energy_internal_mol = pyo.Expression(
                expr=self.u_sp_func(cmp, self.s_kJ_per_kgK, self.p_kPa, _data_dir)
                * params.uc["kJ/kg to J/mol"]
            )
            self.cp_mass = pyo.Expression(
                expr=self.cp_sp_func(cmp, self.s_kJ_per_kgK, self.p_kPa, _data_dir)
                * params.uc["kJ/kg/K to J/kg/K"]
            )
            self.cp_mol = pyo.Expression(
                expr=self.cp_sp_func(cmp, self.s_kJ_per_kgK, self.p_kPa, _data_dir)
                * params.uc["kJ/kg/K to J/mol/K"]
            )
            self.cv_mass = pyo.Expression(
                expr=self.cv_sp_func(cmp, self.s_kJ_per_kgK, self.p_kPa, _data_dir)
                * params.uc["kJ/kg/K to J/kg/K"]
            )
            self.cv_mol = pyo.Expression(
                expr=self.cv_sp_func(cmp, self.s_kJ_per_kgK, self.p_kPa, _data_dir)
                * params.uc["kJ/kg/K to J/mol/K"]
            )
            self.dens_mass = pyo.Expression(
                expr=1.0 / self.v_sp_func(cmp, self.s_kJ_per_kgK, self.p_kPa, _data_dir)
            )
            self.dens_mol = pyo.Expression(
                expr=params.uc["kg/m3 to mol/m3"]
                / self.v_sp_func(cmp, self.s_kJ_per_kgK, self.p_kPa, _data_dir)
            )
        # Internal energy is a state variable
        elif self.state_vars == StateVars.PU:
            if self.amount_basis == AmountBasis.MOLE:
                self.energy_internal_mass = pyo.Expression(
                    expr=self.energy_internal_mol * params.uc["J/mol to J/kg"]
                )
            else:
                self.energy_internal_mol = pyo.Expression(
                    expr=self.energy_internal_mass * params.uc["J/kg to J/mol"]
                )
            self.enth_mass = pyo.Expression(
                expr=self.h_up_func(cmp, self.u_kJ_per_kg, self.p_kPa, _data_dir)
                * params.uc["kJ/kg to J/kg"]
            )
            self.enth_mol = pyo.Expression(
                expr=self.h_up_func(cmp, self.u_kJ_per_kg, self.p_kPa, _data_dir)
                * params.uc["kJ/kg to J/mol"]
            )
            self.entr_mass = pyo.Expression(
                expr=self.s_up_func(cmp, self.u_kJ_per_kg, self.p_kPa, _data_dir)
                * params.uc["kJ/kg/K to J/kg/K"]
            )
            self.entr_mol = pyo.Expression(
                expr=self.s_up_func(cmp, self.u_kJ_per_kg, self.p_kPa, _data_dir)
                * params.uc["kJ/kg/K to J/mol/K"]
            )
            self.cp_mass = pyo.Expression(
                expr=self.cp_up_func(cmp, self.u_kJ_per_kg, self.p_kPa, _data_dir)
                * params.uc["kJ/kg/K to J/kg/K"]
            )
            self.cp_mol = pyo.Expression(
                expr=self.cp_up_func(cmp, self.u_kJ_per_kg, self.p_kPa, _data_dir)
                * params.uc["kJ/kg/K to J/mol/K"]
            )
            self.cv_mass = pyo.Expression(
                expr=self.cv_up_func(cmp, self.u_kJ_per_kg, self.p_kPa, _data_dir)
                * params.uc["kJ/kg/K to J/kg/K"]
            )
            self.cv_mol = pyo.Expression(
                expr=self.cv_up_func(cmp, self.u_kJ_per_kg, self.p_kPa, _data_dir)
                * params.uc["kJ/kg/K to J/mol/K"]
            )
            self.dens_mass = pyo.Expression(
                expr=1.0 / self.v_up_func(cmp, self.u_kJ_per_kg, self.p_kPa, _data_dir)
            )
            self.dens_mol = pyo.Expression(
                expr=params.uc["kg/m3 to mol/m3"]
                / self.v_up_func(cmp, self.u_kJ_per_kg, self.p_kPa, _data_dir)
            )
        else:  # T, P, x
            # enthalpy
            self.enth_mol = pyo.Expression(
                expr=sum(self.phase_frac[p] * self.enth_mol_phase[p] for p in phlist)
            )

            self.enth_mass = pyo.Expression(
                expr=sum(self.phase_frac[p] * self.enth_mass_phase[p] for p in phlist)
            )

            # Entropy
            self.entr_mol = pyo.Expression(
                expr=sum(self.phase_frac[p] * self.entr_mol_phase[p] for p in phlist)
            )
            self.entr_mass = pyo.Expression(
                expr=sum(self.phase_frac[p] * self.entr_mass_phase[p] for p in phlist)
            )
            # Internal Energy
            self.energy_internal_mol = pyo.Expression(
                expr=sum(
                    self.phase_frac[p] * self.energy_internal_mol_phase[p]
                    for p in phlist
                )
            )
            self.energy_internal_mass = pyo.Expression(
                expr=sum(
                    self.phase_frac[p] * self.energy_internal_mass_phase[p]
                    for p in phlist
                )
            )
            # cp
            self.cp_mol = pyo.Expression(
                expr=sum(self.phase_frac[p] * self.cp_mol_phase[p] for p in phlist)
            )
            self.cp_mass = pyo.Expression(
                expr=sum(self.phase_frac[p] * self.cp_mass_phase[p] for p in phlist)
            )
            # cv
            self.cv_mol = pyo.Expression(
                expr=sum(self.phase_frac[p] * self.cv_mol_phase[p] for p in phlist)
            )
            self.cv_mass = pyo.Expression(
                expr=sum(self.phase_frac[p] * self.cv_mass_phase[p] for p in phlist)
            )
            # mass density
            self.dens_mass = pyo.Expression(
                expr=1.0
                / sum(
                    self.phase_frac[p] * 1.0 / self.dens_mass_phase[p] for p in phlist
                )
            )
            # mole density
            self.dens_mol = pyo.Expression(
                expr=1.0
                / sum(self.phase_frac[p] * 1.0 / self.dens_mol_phase[p] for p in phlist)
            )

        # heat capacity ratio
        self.heat_capacity_ratio = pyo.Expression(expr=self.cp_mol / self.cv_mol)
        # Volumetric flow
        self.flow_vol = pyo.Expression(
            expr=self.flow_mol / self.dens_mol,
            doc="Total liquid + vapor volumetric flow",
        )

        def rule_mole_frac_phase_comp(b, p, j):
            if p == "Mix":
                return 1.0
            else:
                return self.phase_frac[p]

        self.mole_frac_phase_comp = pyo.Expression(
            pub_phlist, component_list, rule=rule_mole_frac_phase_comp
        )
        self.mass_frac_phase_comp = pyo.Expression(  # since pure comp same as mole
            pub_phlist, component_list, rule=rule_mole_frac_phase_comp
        )

        if viscosity_available(cmp):

            def rule_visc_d_phase(b, p):
                if p == "Liq":
                    return self.expression_writer.viscosity_liq(
                        **sv_dict_liq, convert_args=False
                    )
                else:
                    return self.expression_writer.viscosity_vap(
                        **sv_dict_vap, convert_args=False
                    )

            self.visc_d_phase = pyo.Expression(
                phlist,
                rule=rule_visc_d_phase,
                doc="(Dynamic) viscosity of phase",
            )

            def rule_visc_k_phase(b, p):
                if p == "Liq":
                    return (
                        self.expression_writer.viscosity_liq(
                            **sv_dict_liq, convert_args=False
                        )
                        / self.dens_mass_phase[p]
                    )
                else:
                    return (
                        self.expression_writer.viscosity_vap(
                            **sv_dict_vap, convert_args=False
                        )
                        / self.dens_mass_phase[p]
                    )

            self.visc_k_phase = pyo.Expression(
                phlist,
                rule=rule_visc_k_phase,
                doc="Kinematic viscosity of phase",
            )

        if thermal_conductivity_available(cmp):

            def rule_therm_cond_phase(b, p):
                if p == "Liq":
                    return self.expression_writer.thermal_conductivity_liq(
                        **sv_dict_liq, convert_args=False
                    )
                else:
                    return self.expression_writer.thermal_conductivity_vap(
                        **sv_dict_vap, convert_args=False
                    )

            self.therm_cond_phase = pyo.Expression(
                phlist,
                rule=rule_therm_cond_phase,
                doc="Thermal conductivity of phase",
            )

        if surface_tension_available(cmp):
            self.surf_tens = pyo.Expression(
                phlist,
                expr=self.expression_writer.surface_tension(
                    **sv_dict_liq, convert_args=False
                ),
                doc="Thermal conductivity of phase",
            )

        # Define some expressions for the balance terms returned by functions
        # This is just to allow assigning scale factors to the expressions
        # returned
        #
        # Material flow term expressions
        if self.amount_basis == AmountBasis.MOLE:

            def rule_material_flow_terms(b, p):
                if p == "Mix":
                    return self.flow_mol
                else:
                    return self.flow_mol * self.phase_frac[p]

        else:

            def rule_material_flow_terms(b, p):
                if p == "Mix":
                    return self.flow_mass
                else:
                    return self.flow_mass * self.phase_frac[p]

        self.material_flow_terms = pyo.Expression(
            pub_phlist, rule=rule_material_flow_terms
        )

        # Enthalpy flow term expressions
        if self.amount_basis == AmountBasis.MOLE:

            def rule_enthalpy_flow_terms(b, p):
                if p == "Mix":
                    return self.enth_mol * self.flow_mol
                else:
                    return self.enth_mol_phase[p] * self.phase_frac[p] * self.flow_mol

        else:

            def rule_enthalpy_flow_terms(b, p):
                if p == "Mix":
                    return self.enth_mass * self.flow_mass
                else:
                    return self.enth_mass_phase[p] * self.phase_frac[p] * self.flow_mass

        self.enthalpy_flow_terms = pyo.Expression(
            pub_phlist, rule=rule_enthalpy_flow_terms
        )

        # Energy density term expressions
        if self.amount_basis == AmountBasis.MOLE:

            def rule_energy_density_terms(b, p):
                if p == "Mix":
                    return self.dens_mol * self.energy_internal_mol
                else:
                    return self.dens_mol_phase[p] * self.energy_internal_mol_phase[p]

        else:

            def rule_energy_density_terms(b, p):
                if p == "Mix":
                    return self.dens_mass * self.energy_internal_mass
                else:
                    return self.dens_mass_phase[p] * self.energy_internal_mass_phase[p]

        self.energy_density_terms = pyo.Expression(
            pub_phlist, rule=rule_energy_density_terms
        )

    def get_material_flow_terms(self, p, j):
        """Get material flow terms for phase

        Args:
            p (str): phase

        Returns:
            Expression
        """
        return self.material_flow_terms[p]

    def get_enthalpy_flow_terms(self, p):
        """Get enthalpy flow terms for phase

        Args:
            p (str): phase

        Returns:
            Expression
        """
        return self.enthalpy_flow_terms[p]

    def get_material_density_terms(self, p, j):
        """Get material density terms for phase

        Args:
            p (str): phase

        Returns:
            Expression
        """
        if self.amount_basis == AmountBasis.MOLE:
            if p == "Mix":
                return self.dens_mol
            else:
                return self.dens_mol_phase[p]
        else:
            if p == "Mix":
                return self.dens_mass
            else:
                return self.dens_mass_phase[p]

    def get_energy_density_terms(self, p):
        """Get energy density terms for phase

        Args:
            p (str): phase

        Returns:
            Expression
        """
        return self.energy_density_terms[p]

    def default_material_balance_type(self):
        """Get default material balance type suggestion"""
        return MaterialBalanceType.componentTotal

    def default_energy_balance_type(self):
        """Get default energy balance type suggestion"""
        return EnergyBalanceType.enthalpyTotal

    def get_material_flow_basis(b):
        return MaterialFlowBasis.molar

    def define_state_vars(self):
        return self._state_vars_dict

    def define_display_vars(self):
        if self.amount_basis == AmountBasis.MOLE:
            return {
                "Molar Flow": self.flow_mol,
                "Mass Flow": self.flow_mass,
                "T": self.temperature,
                "P": self.pressure,
                "Vapor Fraction": self.vapor_frac,
                "Molar Enthalpy": self.enth_mol,
            }
        else:
            return {
                "Molar Flow": self.flow_mol,
                "Mass Flow": self.flow_mass,
                "T": self.temperature,
                "P": self.pressure,
                "Vapor Fraction": self.vapor_frac,
                "Mass Enthalpy": self.enth_mass,
            }

    def extensive_state_vars(self):
        """Return the set of extensive variables"""
        return self.extensive_set

    def intensive_state_vars(self):
        """Return the set of intensive variables"""
        return self.intensive_set

    def model_check(self):
        """Currently doesn't do anything, here for compatibility"""

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        if self.params.config.amount_basis == AmountBasis.MOLE:
            sf_flow = iscale.get_scaling_factor(self.flow_mol, default=1)
            sf_enth = iscale.get_scaling_factor(self.enth_mol, default=1)
            sf_inte = iscale.get_scaling_factor(self.energy_internal_mol, default=1)
            sf_dens = iscale.get_scaling_factor(self.dens_mol, default=1)
        else:
            sf_flow = iscale.get_scaling_factor(self.flow_mass, default=1)
            sf_enth = iscale.get_scaling_factor(self.enth_mass, default=1)
            sf_inte = iscale.get_scaling_factor(self.energy_internal_mass, default=1)
            sf_dens = iscale.get_scaling_factor(self.dens_mass, default=1)
        sf_pres = iscale.get_scaling_factor(self.pressure, default=1)

        for v in self.material_flow_terms.values():
            iscale.set_scaling_factor(v, sf_flow)
        for v in self.enthalpy_flow_terms.values():
            iscale.set_scaling_factor(v, sf_enth * sf_flow)
        for k, v in self.energy_density_terms.items():
            if k == "Mix":
                iscale.set_scaling_factor(v, sf_inte * sf_dens)
            else:
                if self.params.config.amount_basis == AmountBasis.MOLE:
                    sf_inte_p = iscale.get_scaling_factor(
                        self.energy_internal_mol_phase[k], default=1
                    )
                    sf_dens_p = iscale.get_scaling_factor(
                        self.dens_mol_phase[k], default=1
                    )
                else:
                    sf_inte_p = iscale.get_scaling_factor(
                        self.energy_internal_mass_phase[k], default=1
                    )
                    sf_dens_p = iscale.get_scaling_factor(
                        self.dens_mass_phase[k], default=1
                    )
                iscale.set_scaling_factor(v, sf_inte_p * sf_dens_p)
        try:
            iscale.set_scaling_factor(self.eq_sat, sf_pres / 1000.0)
        except AttributeError:
            pass  # may not have eq_sat, and that's ok
        try:
            iscale.set_scaling_factor(self.eq_complementarity, sf_pres / 10)
        except AttributeError:
            pass  # may not have eq_complementarity which is fine



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import pytest

import pyomo.environ as pyo
from pyomo.common import unittest as pyo_unittest

from idaes.apps.grid_integration.bidder import Bidder
from idaes.apps.grid_integration.tests.util import (
    ExampleModel,
    ExampleForecaster,
    testing_model_data,
)
from idaes.apps.grid_integration.coordinator import prescient_avail


class MissingModel:
    """
    A class for testing missing methods and attributes.
    """

    method_dict = {k: lambda: None for k in ["populate_model", "update_model"]}
    attr_dict = {
        "power_output": "power_output",
        "total_cost": ("tot_cost", 1),
        "generator": "test",
        "pmin": 20.00,
        "default_bids": {p: 30 for p in [20.00, 40.00, 60.00, 80.00, 100.00]},
    }

    def __init__(self, missing_method=None, missing_attr=None):
        """
        Constructs a model class without the specified missing methods and/or
        missing attributes.
        """

        for m in self.method_dict:
            if m != missing_method:
                setattr(self, m, self.method_dict[m])

        for a in self.attr_dict:
            if a != missing_attr:
                setattr(self, a, self.attr_dict[a])


horizon = 4
n_scenario = 3

day_ahead_horizon = horizon
real_time_horizon = horizon


@pytest.mark.unit
def test_model_object_missing_methods():

    solver = pyo.SolverFactory("cbc")
    forecaster = ExampleForecaster(prediction=30)

    # By definition, the model object should contain these methods
    method_list = ["populate_model", "update_model"]

    # test if the correct error message is raised if a model misses necessary methods
    for m in method_list:
        bidding_model_object = MissingModel(missing_method=m)
        with pytest.raises(AttributeError, match=r".*{}().*".format(m)):
            bidder_object = Bidder(
                bidding_model_object=bidding_model_object,
                day_ahead_horizon=day_ahead_horizon,
                real_time_horizon=real_time_horizon,
                n_scenario=n_scenario,
                solver=solver,
                forecaster=forecaster,
            )


@pytest.mark.unit
def test_model_object_missing_attr():

    solver = pyo.SolverFactory("cbc")
    forecaster = ExampleForecaster(prediction=30)

    # By definition, the model object should contain these attributes
    attr_list = ["power_output", "total_cost", "model_data"]

    # test if the correct error message is raised if a model misses necessary attributes
    for attr in attr_list:
        bidding_model_object = MissingModel(missing_attr=attr)
        with pytest.raises(AttributeError, match=r".*{}().*".format(attr)):
            bidder_object = Bidder(
                bidding_model_object=bidding_model_object,
                day_ahead_horizon=day_ahead_horizon,
                real_time_horizon=real_time_horizon,
                n_scenario=n_scenario,
                solver=solver,
                forecaster=forecaster,
            )


@pytest.mark.unit
def test_n_scenario_checker():

    solver = pyo.SolverFactory("cbc")
    forecaster = ExampleForecaster(prediction=30)
    bidding_model_object = ExampleModel(model_data=testing_model_data)

    # test if bidder raise error when negative number of scenario is given
    with pytest.raises(ValueError, match=r".*greater than zero.*"):
        bidder_object = Bidder(
            bidding_model_object=bidding_model_object,
            day_ahead_horizon=day_ahead_horizon,
            real_time_horizon=real_time_horizon,
            n_scenario=-1,
            solver=solver,
            forecaster=forecaster,
        )

    # test if bidder raise error when float number of scenario is given
    with pytest.raises(TypeError, match=r".*should be an integer.*"):
        bidder_object = Bidder(
            bidding_model_object=bidding_model_object,
            day_ahead_horizon=day_ahead_horizon,
            real_time_horizon=real_time_horizon,
            n_scenario=3.0,
            solver=solver,
            forecaster=forecaster,
        )


@pytest.mark.unit
def test_solver_checker():

    forecaster = ExampleForecaster(prediction=30)
    bidding_model_object = ExampleModel(model_data=testing_model_data)

    # test if bidder raise error when invalid solver is provided
    invalid_solvers = [5, "cbc", "ipopt"]
    for s in invalid_solvers:
        with pytest.raises(TypeError, match=r".*not a valid Pyomo solver.*"):
            bidder_object = Bidder(
                bidding_model_object=bidding_model_object,
                day_ahead_horizon=day_ahead_horizon,
                real_time_horizon=real_time_horizon,
                n_scenario=n_scenario,
                solver=s,
                forecaster=forecaster,
            )


@pytest.fixture
def bidder_object():

    solver = pyo.SolverFactory("cbc")
    forecaster = ExampleForecaster(prediction=30)

    # create a bidder model
    bidding_model_object = ExampleModel(model_data=testing_model_data)
    bidder_object = Bidder(
        bidding_model_object=bidding_model_object,
        day_ahead_horizon=day_ahead_horizon,
        real_time_horizon=real_time_horizon,
        n_scenario=n_scenario,
        solver=solver,
        forecaster=forecaster,
    )
    return bidder_object


@pytest.mark.component
@pytest.mark.skipif(
    not prescient_avail, reason="Prescient (optional dependency) not available"
)
def test_compute_DA_bids(bidder_object):

    marginal_cost = bidder_object.bidding_model_object.marginal_cost
    gen = bidder_object.generator
    default_bids = bidder_object.bidding_model_object.model_data.p_cost
    pmin = bidder_object.bidding_model_object.pmin
    pmax = bidder_object.bidding_model_object.pmax
    date = "2021-08-20"

    # test bidding when price forecast lower than marginal cost
    # expect default bids
    shift = 1
    fixed_forecast = marginal_cost - shift
    bidder_object.forecaster.prediction = fixed_forecast
    bids = bidder_object.compute_day_ahead_bids(date=date, hour=0)

    expected_bids = {
        t: {
            gen: {
                "p_min": pmin,
                "p_max": pmax,
                "p_min_agc": pmin,
                "p_max_agc": pmax,
                "startup_capacity": pmin,
                "shutdown_capacity": pmin,
                "p_cost": [
                    (p, p * marginal_cost - shift * pmin) for p, _ in default_bids
                ],
            }
        }
        for t in range(horizon)
    }

    pyo_unittest.assertStructuredAlmostEqual(first=expected_bids, second=bids)

    # test bidding when price forecast higher than marginal cost
    shift = 1
    fixed_forecast = marginal_cost + shift
    bidder_object.forecaster.prediction = fixed_forecast
    bids = bidder_object.compute_day_ahead_bids(date=date, hour=0)

    expected_bids = {}
    for t in range(horizon):

        expected_bids[t] = {}
        expected_bids[t][gen] = {
            "p_min": pmin,
            "p_max": pmax,
            "p_min_agc": pmin,
            "p_max_agc": pmax,
            "startup_capacity": pmin,
            "shutdown_capacity": pmin,
        }
        p_cost = []

        pre_p = 0
        pre_cost = 0

        for p, _ in default_bids:
            # because the price forecast is higher than the marginal costs
            # to have highest profit, power output will be pmax
            # and the bidding costs will be computed with the price forecasts
            if p == pmax:
                p_cost.append((p, (p - pre_p) * (marginal_cost + shift) + pre_cost))
            else:
                p_cost.append((p, (p - pre_p) * marginal_cost + pre_cost))
            pre_p = p
            pre_cost = p_cost[-1][1]

        expected_bids[t][gen]["p_cost"] = p_cost

    pyo_unittest.assertStructuredAlmostEqual(first=expected_bids, second=bids)


@pytest.mark.component
@pytest.mark.skipif(
    not prescient_avail, reason="Prescient (optional dependency) not available"
)
def test_compute_RT_bids(bidder_object):
    marginal_cost = bidder_object.bidding_model_object.marginal_cost
    gen = bidder_object.generator
    default_bids = bidder_object.bidding_model_object.model_data.p_cost
    pmin = bidder_object.bidding_model_object.pmin
    pmax = bidder_object.bidding_model_object.pmax
    date = "2021-08-20"

    # test bidding when price forecast lower than marginal cost
    # expect default bids
    realized_day_ahead_prices = [29] * bidder_object.real_time_horizon
    realized_day_ahead_dispatches = [0] * bidder_object.real_time_horizon
    shift = 1
    fixed_forecast = marginal_cost - shift
    bidder_object.forecaster.prediction = fixed_forecast
    bids = bidder_object.compute_real_time_bids(
        date=date,
        hour=0,
        realized_day_ahead_prices=realized_day_ahead_prices,
        realized_day_ahead_dispatches=realized_day_ahead_dispatches,
    )

    expected_bids = {
        t: {
            gen: {
                "p_min": pmin,
                "p_max": pmax,
                "p_min_agc": pmin,
                "p_max_agc": pmax,
                "startup_capacity": pmin,
                "shutdown_capacity": pmin,
                "p_cost": [(p, (p - pmin) * marginal_cost) for p, _ in default_bids],
            }
        }
        for t in range(horizon)
    }

    pyo_unittest.assertStructuredAlmostEqual(first=expected_bids, second=bids)


@pytest.fixture
def bidder_object_pcost():

    solver = pyo.SolverFactory("cbc")
    forecaster = ExampleForecaster(prediction=30)

    # create a bidder model
    testing_model_data.include_default_p_cost = False
    bidding_model_object = ExampleModel(model_data=testing_model_data)
    bidder_object = Bidder(
        bidding_model_object=bidding_model_object,
        day_ahead_horizon=day_ahead_horizon,
        real_time_horizon=real_time_horizon,
        n_scenario=n_scenario,
        solver=solver,
        forecaster=forecaster,
    )
    return bidder_object


@pytest.mark.component
@pytest.mark.skipif(
    not prescient_avail, reason="Prescient (optional dependency) not available"
)
def test_compute_RT_bids_pcost(bidder_object_pcost):
    marginal_cost = bidder_object_pcost.bidding_model_object.marginal_cost
    gen = bidder_object_pcost.generator
    default_bids = bidder_object_pcost.bidding_model_object.model_data.p_cost
    pmin = bidder_object_pcost.bidding_model_object.pmin
    pmax = bidder_object_pcost.bidding_model_object.pmax
    date = "2021-08-20"

    # test bidding when price forecast lower than marginal cost
    # expect default bids
    realized_day_ahead_prices = [29] * bidder_object_pcost.real_time_horizon
    realized_day_ahead_dispatches = [0] * bidder_object_pcost.real_time_horizon
    shift = 1
    fixed_forecast = marginal_cost - shift
    bidder_object_pcost.forecaster.prediction = fixed_forecast
    bids = bidder_object_pcost.compute_real_time_bids(
        date=date,
        hour=0,
        realized_day_ahead_prices=realized_day_ahead_prices,
        realized_day_ahead_dispatches=realized_day_ahead_dispatches,
    )

    expected_bids = {
        t: {
            gen: {
                "p_min": pmin,
                "p_max": pmin,
                "p_min_agc": pmin,
                "p_max_agc": pmin,
                "startup_capacity": pmin,
                "shutdown_capacity": pmin,
                "p_cost": [(30, 0)],
            }
        }
        for t in range(horizon)
    }

    pyo_unittest.assertStructuredAlmostEqual(first=expected_bids, second=bids)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Utility functions for column models
"""

__author__ = "Jaffer Ghouse, Alejandro Garcia-Diego"

# TODO: look into protected access issues - probably need to refactor
# pylint: disable=protected-access

from functools import partial

# Import Pyomo libraries
from pyomo.environ import (
    Reference,
    Expression,
)

# Import IDAES cores
from idaes.core.util.exceptions import (
    PropertyPackageError,
    PropertyNotSupportedError,
)
import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)


def make_phase_split(
    model,
    port=None,
    phase=None,
    side_sf=None,
    equipmentType=None,
):
    """Method to split and populate the outlet ports with corresponding
    phase values from the mixed stream outlet block."""
    t0 = model.flowsheet().time.first()

    member_list = model.properties_out[t0].define_port_members()

    for k in member_list:

        local_name = member_list[k].local_name

        # Create references and populate the intensive variables
        if (
            "flow" not in local_name
            and "frac" not in local_name
            and "enth" not in local_name
        ):
            if not member_list[k].is_indexed():
                var = model.properties_out[:].component(local_name)
            else:
                var = model.properties_out[:].component(local_name)[...]

            # add the reference and variable name to the port
            port.add(Reference(var), k)

        elif "frac" in local_name:

            # Mole/mass frac is typically indexed
            index_set = member_list[k].index_set()

            # if state var is not mole/mass frac by phase
            if "phase" not in local_name:
                if "mole" in local_name:  # check mole basis/mass basis

                    # The following conditionals are required when a
                    # mole frac or mass frac is a state var i.e. will be
                    # a port member. This gets a bit tricky when handling
                    # non-conventional systems when you have more than one
                    # liquid or vapor phase. Hence, the logic here is that
                    # the mole frac that should be present in the liquid or
                    # vapor port should be computed by accounting for
                    # multiple liquid or vapor phases if present. For the
                    # classical VLE system, this holds too.
                    if hasattr(
                        model.properties_out[t0], "mole_frac_phase_comp"
                    ) and hasattr(model.properties_out[t0], "flow_mol_phase"):
                        flow_phase_comp = False
                        local_name_frac = "mole_frac_phase_comp"
                        local_name_flow = "flow_mol_phase"
                    elif hasattr(model.properties_out[t0], "flow_mol_phase_comp"):
                        flow_phase_comp = True
                        local_name_flow = "flow_mol_phase_comp"
                    else:
                        raise PropertyNotSupportedError(
                            "No mole_frac_phase_comp or flow_mol_phase or"
                            " flow_mol_phase_comp variables encountered "
                            "while building ports. "
                        )
                elif "mass" in local_name:
                    if hasattr(
                        model.properties_out[t0], "mass_frac_phase_comp"
                    ) and hasattr(model.properties_out[t0], "flow_mass_phase"):
                        flow_phase_comp = False
                        local_name_frac = "mass_frac_phase_comp"
                        local_name_flow = "flow_mass_phase"
                    elif hasattr(model.properties_out[t0], "flow_mass_phase_comp"):
                        flow_phase_comp = True
                        local_name_flow = "flow_mass_phase_comp"
                    else:
                        raise PropertyNotSupportedError(
                            "No mass_frac_phase_comp or flow_mass_phase or"
                            " flow_mass_phase_comp variables encountered "
                            "while building ports."
                        )
                else:
                    raise PropertyNotSupportedError(
                        "No mass frac or mole frac variables encountered "
                        " while building ports. "
                        "phase_frac as a state variable is not "
                        "supported with distillation unit models."
                    )

                if not flow_phase_comp:
                    tmp_rule = partial(
                        _rule_mole_frac_0,
                        local_name_frac=local_name_frac,
                        local_name_flow=local_name_flow,
                        phase=phase,
                    )
                else:
                    tmp_rule = partial(
                        _rule_mole_frac_1,
                        local_name_flow=local_name_flow,
                        phase=phase,
                    )
                # add the reference and variable name to the port
                expr = Expression(model.flowsheet().time, index_set, rule=tmp_rule)
                model.add_component("e_mole_frac_" + port.local_name, expr)
                port.add(expr, k)
            else:

                # Assumes mole_frac_phase or mass_frac_phase exist as
                # state vars in the port and therefore access directly
                # from the state block.
                var = model.properties_out[:].component(local_name)[...]

                # add the reference and variable name to the port
                ref = Reference(var)
                setattr(model, "_" + k + "_" + port.local_name + "_ref", ref)
                port.add(ref, k)
        elif "flow" in local_name:
            if "phase" not in local_name:

                # Assumes that here the var is total flow or component
                # flow. However, need to extract the flow by phase from
                # the state block. Expects to find the var
                # flow_mol_phase or flow_mass_phase in the state block.

                # Check if it is not indexed by component list and this
                # is total flow
                if "total" in str(equipmentType):
                    if not member_list[k].is_indexed():

                        # add the reference and variable name to the port
                        expr = Expression(
                            model.flowsheet().time,
                            rule=partial(
                                _rule_flow_0, local_name=local_name, side_sf=side_sf
                            ),
                        )
                        model.add_component("e_flow_" + port.local_name, expr)
                        port.add(expr, k)
                    else:
                        # Create references and populate the extensive variables
                        # This is for vars that are indexed by phase, comp or both.
                        index_set = member_list[k].index_set()

                        # add the reference and variable name to the port
                        expr = Expression(
                            model.flowsheet().time,
                            index_set,
                            rule=partial(
                                _rule_flow_1, local_name=local_name, side_sf=side_sf
                            ),
                        )
                        model.add_component("e_flow_" + port.local_name, expr)
                        port.add(expr, k)

                else:
                    if not member_list[k].is_indexed():
                        # if state var is not flow_mol/flow_mass
                        # by phase
                        local_name_flow = local_name + "_phase"

                        # add the reference and variable name to the port
                        expr = Expression(
                            model.flowsheet().time,
                            rule=partial(
                                _rule_flow_2,
                                local_name_flow=local_name_flow,
                                side_sf=side_sf,
                                phase=phase,
                            ),
                        )
                        model.add_component("e_flow_" + port.local_name, expr)
                        port.add(expr, k)
                    else:
                        # when it is flow comp indexed by component list
                        str_split = local_name.split("_")
                        if len(str_split) == 3 and str_split[-1] == "comp":
                            local_name_flow = (
                                str_split[0] + "_" + str_split[1] + "_phase_" + "comp"
                            )

                        # Get the indexing set i.e. component list
                        index_set = member_list[k].index_set()

                        expr = Expression(
                            model.flowsheet().time,
                            index_set,
                            rule=partial(
                                _rule_flow_3,
                                local_name_flow=local_name_flow,
                                side_sf=side_sf,
                                phase=phase,
                            ),
                        )
                        model.add_component("e_flow_" + port.local_name, expr)
                        port.add(expr, local_name)
            else:
                # flow is indexed by phase and comp
                # Get the indexing sets i.e. component list and phase list
                component_set = model.config.property_package.component_list

                phase_set = model.config.property_package.phase_list

                expr = Expression(
                    model.flowsheet().time,
                    phase_set,
                    component_set,
                    rule=partial(
                        _rule_flow_4,
                        local_name=local_name,
                        side_sf=side_sf,
                        phase=phase,
                        equipmentType=equipmentType,
                    ),
                )

                model.add_component("e_" + local_name + port.local_name, expr)
                port.add(expr, k)

        elif "enth" in local_name:
            if "phase" not in local_name:
                # assumes total mixture enthalpy (enth_mol or enth_mass)
                if not member_list[k].is_indexed():
                    # if state var is not enth_mol/enth_mass
                    # by phase, add _phase string to extract the right
                    # value from the state block
                    local_name_phase = local_name + "_phase"
                else:
                    raise PropertyPackageError(
                        "Enthalpy is indexed but the variable "
                        "name does not reflect the presence of an index. "
                        "Please follow the naming convention outlined "
                        "in the documentation for state variables."
                    )

                expr = Expression(
                    model.flowsheet().time,
                    rule=partial(
                        _rule_enth_0, local_name_phase=local_name_phase, phase=phase
                    ),
                )
                model.add_component("e_enth_" + port.local_name, expr)
                # add the reference and variable name to the port
                port.add(expr, k)

            else:
                # assumes enth_mol_phase or enth_mass_phase.
                # This is an intensive property, you create a direct
                # reference irrespective of the reflux, distillate and
                # vap_outlet

                if not member_list[k].is_indexed():
                    var = model.properties_out[:].component(local_name)
                else:
                    var = model.properties_out[:].component(local_name)[...]

                # add the reference and variable name to the port
                ref = Reference(var)
                setattr(model, "_" + k + "_" + port.local_name + "_ref", ref)
                port.add(ref, k)


# Rule for mole fraction
def _rule_mole_frac_0(model, t, i, local_name_frac, local_name_flow, phase):
    sum_flow_comp = sum(
        model.properties_out[t].component(local_name_frac)[p, i]
        * model.properties_out[t].component(local_name_flow)[p]
        for p in phase
    )

    return sum_flow_comp / sum(
        model.properties_out[t].component(local_name_flow)[p] for p in phase
    )


# Rule for mole fraction
def _rule_mole_frac_1(model, t, i, local_name_flow, phase):
    sum_flow_comp = sum(
        model.properties_out[t].component(local_name_flow)[p, i] for p in phase
    )

    return sum_flow_comp / sum(
        model.properties_out[t].component(local_name_flow)[p, i]
        for p in phase
        for i in model.config.property_package.component_list
    )


def _rule_flow_0(model, t, local_name, side_sf):
    return (model.properties_out[t].component(local_name)) * (side_sf)


# Rule to link the flow to the port
def _rule_flow_1(model, t, i, local_name, side_sf):
    return (model.properties_out[t].component(local_name)[i]) * (side_sf)


# Rule to link the flow to the port
def _rule_flow_2(model, t, local_name_flow, side_sf, phase):
    return sum(model.properties_out[t].component(local_name_flow)[p] for p in phase) * (
        side_sf
    )


# Rule to link the flow to the port
def _rule_flow_3(model, t, i, local_name_flow, side_sf, phase):
    return sum(
        model.properties_out[t].component(local_name_flow)[p, i] for p in phase
    ) * (side_sf)


def _rule_flow_4(model, t, p, i, local_name, side_sf, phase, equipmentType):
    # If statement to skip in case equipment is not a Tray
    if equipmentType is None:
        if (phase is model._liquid_set and p in model._liquid_set) or (
            phase is model._vapor_set and p in model._vapor_set
        ):
            # pass appropriate phase flow values to port
            return (model.properties_out[t].component(local_name)[p, i]) * (side_sf)

        else:
            # return small number for phase that should not
            # be in the appropriate port. For example,
            # the state vars will be flow_mol_phase_comp
            # which will include all phases. The liq port
            # should have the correct references to the liq
            # phase flow but the vapor phase flow should be 0.
            # TODO this should be a param and probably also have units
            return 1e-8
    else:
        return (model.properties_out[t].component(local_name)[p, i]) * (side_sf)


# Rule to link the phase enthalpy to the port.
def _rule_enth_0(model, t, local_name_phase, phase):
    return sum(model.properties_out[t].component(local_name_phase)[p] for p in phase)
